import json
import numpy as np


def _create_classes(classes_map):
    classes_loader = []
    for key in classes_map:
        color = np.random.choice(range(256), size=3)
        hexcolor = "#%02x%02x%02x" % tuple(color)
        sa_classes = {'name': key, 'color': hexcolor, 'attribute_groups': []}
        classes_loader.append(sa_classes)
    return classes_loader


def vott_object_detection_to_sa_vector(file_list):
    sa_jsons = {}
    classes = []
    for json_file in file_list:
        json_data = json.load(open(json_file))
        file_name = json_data['asset']['name'] + '___objects.json'

        instances = json_data['regions']
        sa_loader = []
        for instance in instances:
            for tag in instance['tags']:
                classes.append(tag)
            sa_obj = {
                'className': instance['tags'][0],
                'attributes': [],
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': 0
            }
            if instance['type'] == 'RECTANGLE' or instance['type'] == 'POLYGON':
                sa_obj['type'] = 'bbox'
                sa_obj['points'] = {
                    'x1':
                        instance['boundingBox']['left'],
                    'y1':
                        instance['boundingBox']['top'],
                    'x2':
                        instance['boundingBox']['left'] +
                        instance['boundingBox']['width'],
                    'y2':
                        instance['boundingBox']['top'] +
                        instance['boundingBox']['height'],
                }
                sa_loader.append(sa_obj.copy())

        sa_jsons[file_name] = sa_loader

    sa_classes = _create_classes(set(classes))
    return sa_jsons, sa_classes


def vott_instance_segmentation_to_sa_vector(file_list):
    sa_jsons = {}
    classes = []
    for json_file in file_list:
        json_data = json.load(open(json_file))
        file_name = json_data['asset']['name'] + '___objects.json'

        instances = json_data['regions']
        sa_loader = []
        for instance in instances:
            classes.append(instance['tags'][0])
            for tag in instance['tags']:
                classes.append(tag)

            sa_obj = {
                'className': instance['tags'][0],
                'attributes': [],
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': 0
            }
            if instance['type'] == 'POLYGON':
                sa_obj['type'] = 'polygon'
                sa_obj['points'] = []
                for point in instance['points']:
                    sa_obj['points'].append(point['x'])
                    sa_obj['points'].append(point['y'])
                sa_loader.append(sa_obj.copy())

        sa_jsons[file_name] = sa_loader

    sa_classes = _create_classes(set(classes))
    return sa_jsons, sa_classes


def vott_to_sa(file_list):
    sa_jsons = {}
    classes = []
    for json_file in file_list:
        json_data = json.load(open(json_file))
        file_name = json_data['asset']['name'] + '___objects.json'

        instances = json_data['regions']
        sa_loader = []
        for instance in instances:
            classes.append(instance['tags'][0])
            for tag in instance['tags']:
                classes.append(tag)

            sa_obj = {
                'className': instance['tags'][0],
                'attributes': [],
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': 0
            }
            if instance['type'] == 'RECTANGLE':
                sa_obj['type'] = 'bbox'
                sa_obj['points'] = {
                    'x1':
                        instance['boundingBox']['left'],
                    'y1':
                        instance['boundingBox']['top'],
                    'x2':
                        instance['boundingBox']['left'] +
                        instance['boundingBox']['width'],
                    'y2':
                        instance['boundingBox']['top'] +
                        instance['boundingBox']['height'],
                }
                sa_loader.append(sa_obj.copy())
            elif instance['type'] == 'POLYGON':
                sa_obj['type'] = 'polygon'
                sa_obj['points'] = []
                for point in instance['points']:
                    sa_obj['points'].append(point['x'])
                    sa_obj['points'].append(point['y'])
                sa_loader.append(sa_obj.copy())

        sa_jsons[file_name] = sa_loader

    sa_classes = _create_classes(set(classes))
    return sa_jsons, sa_classes



import json
import logging

from ....pycocotools_sa import mask as cocomask

logger = logging.getLogger("superannotate-python-sdk")


def sa_vector_to_coco_object_detection(
    make_annotation, image_commons, id_generator, cat_id_map
):
    annotations_per_image = []
    image_info = image_commons.image_info
    sa_ann_json = image_commons.sa_ann_json

    for instance in sa_ann_json:
        if instance['type'] != 'bbox':
            logger.warning(
                "Skipping '{}' type convertion during object_detection task".
                format(instance['type'])
            )
            continue

        anno_id = next(id_generator)
        try:
            if 'className' in instance and instance['className'] in cat_id_map:
                category_id = cat_id_map[instance['className']]
            else:
                category_id = instance['classId']

            points = instance['points']
            for key in points:
                points[key] = round(points[key], 2)
            bbox = (
                points['x1'], points['y1'], points['x2'] - points['x1'],
                points['y2'] - points['y1']
            )
            polygons = bbox
            area = int(
                (points['x2'] - points['x1']) * points['y2'] - points['y1']
            )

            annotation = make_annotation(
                category_id, image_info['id'], bbox, polygons, area, anno_id
            )
            annotations_per_image.append(annotation)
        except Exception as e:
            print(e)

    return image_info, annotations_per_image


def sa_vector_to_coco_instance_segmentation(
    make_annotation, image_commons, id_generator, cat_id_map
):
    annotations_per_image = []
    grouped_polygons = {}

    image_info = image_commons.image_info
    sa_ann_json = image_commons.sa_ann_json
    for instance in sa_ann_json:
        if instance['type'] != 'polygon':
            logger.warning(
                "Skipping '{}' type convertion during object_detection task".
                format(instance['type'])
            )
            continue

        group_id = instance['groupId']
        if 'className' in instance and instance['className'] in cat_id_map:
            category_id = cat_id_map[instance['className']]
        else:
            category_id = instance['classId']

        points = [round(point, 2) for point in instance['points']]
        grouped_polygons.setdefault(group_id, {}).setdefault(category_id,
                                                             []).append(points)

    for polygon_group in grouped_polygons.values():
        for cat_id, polygons in polygon_group.items():
            anno_id = next(id_generator)
            try:
                masks = cocomask.frPyObjects(
                    polygons, image_info['height'], image_info['width']
                )
                mask = cocomask.merge(masks)
                area = int(cocomask.area(mask))
                bbox = cocomask.toBbox(mask).tolist()
                annotation = make_annotation(
                    cat_id, image_info['id'], bbox, polygons, area, anno_id
                )
                annotations_per_image.append(annotation)
            except Exception as e:
                print(e)
    return (image_info, annotations_per_image)


def sa_vector_to_coco_keypoint_detection(
    json_paths, id_generator, id_generator_anno, id_generator_img,
    make_image_info
):
    def __make_skeleton(template):
        res = [
            [connection['from'], connection['to']]
            for connection in template['connections']
        ]
        return res

    def __make_keypoints(template):
        int_dict = {
            int(key): value
            for key, value in template['pointLabels'].items()
        }

        res = [int_dict[i] for i in range(len(int_dict))]
        return res

    def __make_bbox(points):
        xs = [point['x'] for point in points]
        ys = [point['y'] for point in points]

        return [
            int(min(xs)),
            int(min(ys)),
            int(max(xs)) - int(min(xs)),
            int(max(ys)) - int(min(ys))
        ]

    def __load_one_json(path_):
        with open(path_) as fp:
            data = json.load(fp)
        return data

    def __make_annotations(template, id_generator, cat_id, image_id):
        keypoints = []

        for point in template['points']:
            keypoints += [round(point['x'], 2), round(point['y'], 2), 2]
        bbox = __make_bbox(template['points'])
        annotation = {
            'id': next(id_generator),
            'image_id': image_id,
            'iscrowd': 0,
            'bbox': bbox,
            'area': bbox[2] * bbox[3],
            'num_keypoints': len(template['points']),
            'keypoints': keypoints,
            'category_id': cat_id
        }

        return annotation

    template_names = set()
    categories = []
    annotations = []
    images = []

    for path_ in json_paths:
        json_data = __load_one_json(path_)

        for instance in json_data:
            if instance['type'] == 'template' and 'templateId' not in instance:
                logger.warning(
                    'There was a template with no "templateName". \
                                This can happen if the template was deleted from annotate.online. Ignoring this annotation'
                )
                continue

            if instance['type'] != 'template' or instance['templateId'
                                                         ] in template_names:
                continue
            if instance['pointLabels'] == {}:
                instance['pointLabels'] = {
                    x['id']: x['id']
                    for x in instance['points']
                }

            keypoints = []
            name = ""
            supercategory = ""
            id_ = 0
            try:
                template_names.add(str(instance['templateId']))
                skeleton = __make_skeleton(instance)
                keypoints = __make_keypoints(instance)
                id_ = next(id_generator)
                if "classId" in instance.keys():
                    supercategory = instance["classId"]
                else:
                    supercategory = str(instance["templateId"])
                if "className" in instance.keys():
                    name = str(instance["className"])
                else:
                    name = str(instance["templateId"])
            except Exception as e:
                logging.error(e)

            category_item = {
                'name': name,
                'supercategory': supercategory,
                'skeleton': skeleton,
                'keypoints': keypoints,
                'id': id_
            }
            categories.append(category_item)
        image_id = next(id_generator_img)
        image_info = make_image_info(path_, image_id, 'Vector')
        images.append(image_info)

        for instance in json_data:
            cat_id = None
            if instance['type'] == 'template':
                for cat in categories:
                    if cat["name"] == instance["className"]:
                        cat_id = cat['supercategory']

                annotation = __make_annotations(
                    instance, id_generator_anno, cat_id, image_info['id']
                )
                annotations.append(annotation)
    return (categories, annotations, images)



class VocConverter():
    def __init__(self, args):

        self.project_type = args.project_type
        self.dataset_name = args.dataset_name
        self.export_root = args.export_root
        self.output_dir = args.output_dir
        self.task = args.task
        self.direction = args.direction
        self.platform = args.platform



"""
Module which will run converters and convert from
superannotate annotation format to other annotation formats
"""
import copy
import glob
import json
import logging
import os
import shutil

from pathlib import Path

import tempfile
import numpy as np

from .converters.converters import Converter

logger = logging.getLogger("superannotate-python-sdk")


def _split_json(input_dir):
    temp_path = Path(tempfile.mkdtemp()) / 'WebApp'
    temp_path.mkdir()
    json_data = json.load(open(input_dir / "annotations.json"))
    images = json_data.keys()
    for img in images:
        annotations = json_data[img]
        objects = []
        for annot in annotations:
            if annot["type"] != "meta":
                objects.append(annot)
        with open(temp_path / (img + "___objects.json"), "w") as fw:
            json.dump(objects, fw, indent=2)
        shutil.copy(input_dir / "images" / img, temp_path / img)
    (temp_path / "classes").mkdir()
    shutil.copy(
        input_dir / "classes.json", temp_path / "classes" / "classes.json"
    )

    return temp_path


def _load_files(path_to_imgs, task, ptype):
    suffix = None
    if ptype == "Pixel":
        suffix = '___pixel.json'
    else:
        suffix = '___objects.json'

    orig_images = glob.glob(os.path.join(path_to_imgs, '*'))
    orig_images = [
        x for x in orig_images if not os.path.isdir(x) and
        '___' not in x.split('.')[-2] and "mapper.json" not in x
    ]
    all_files = None
    if task == 'keypoint_detection':
        all_files = np.array([(fname, fname + suffix) for fname in orig_images])

    elif ptype == 'Pixel':
        all_files = np.array(
            [
                (fname, fname + suffix, fname + '___save.png')
                for fname in orig_images
            ]
        )
    elif ptype == 'Vector':
        all_files = np.array([(fname, fname + suffix) for fname in orig_images])

    return all_files


def _move_files(data_set, src):
    train_path = src / 'image_set'
    if data_set is not None:
        for tup in data_set:
            for i in tup:
                shutil.copy(i, train_path / Path(i).name)


def _create_classes_mapper(imgs, classes_json):
    classes = {}
    j_data = json.load(open(classes_json))
    for i, instance in enumerate(j_data):
        classes[instance['name']] = i + 1

    with open(imgs / 'image_set' / 'classes_mapper.json', 'w') as fp:
        json.dump(classes, fp)


def export_from_sa(args):
    """
    :param args: All arguments that will be used during convertion.
    :type args: Namespace
    """
    if args.platform == "Desktop":
        args.input_dir = _split_json(args.input_dir)

    data_set = None

    (args.output_dir / "image_set").mkdir(parents=True, exist_ok=True)

    try:
        _create_classes_mapper(
            args.output_dir, args.input_dir / 'classes' / 'classes.json'
        )
    except Exception as e:
        _create_classes_mapper(args.input_dir, args.output_dir)

    data_set = _load_files(args.input_dir, args.task, args.project_type)
    _move_files(data_set, args.output_dir)

    args.__dict__.update(
        {
            'direction': 'to',
            'export_root': args.output_dir / 'image_set'
        }
    )
    converter = Converter(args)

    if data_set is not None:
        converter.strategy.set_dataset_name(args.dataset_name)
        converter.convert_from_sa()

    if args.platform == "Desktop":
        shutil.rmtree(args.input_dir)
    image_set_failed = copy.deepcopy(converter.strategy.failed_conversion_cnt)

    logger.info('Conversion completed successfully')
    return image_set_failed



"""
"""
import json
from collections import namedtuple
from datetime import datetime
from pathlib import Path

import numpy as np
from PIL import Image

from ....common import id2rgb


class CoCoConverter():
    def __init__(self, args):
        self.project_type = args.project_type
        self.dataset_name = args.dataset_name
        self.export_root = args.export_root
        self.output_dir = args.output_dir
        self.task = args.task
        self.direction = args.direction
        self.platform = args.platform

        self.failed_conversion_cnt = 0

    def _create_single_category(self, item):
        category = {
            'id': item.id,
            'name': item.class_name,
            'supercategory': item.class_name,
            'isthing': 1,
            'color': id2rgb(item.id)
        }
        return category

    def set_output_dir(self, output_dir_):
        self.output_dir = output_dir_

    def set_export_root(self, export_dir):
        self.export_root = export_dir

    def set_dataset_name(self, dname):
        self.dataset_name = dname

    def increase_converted_count(self):
        self.failed_conversion_cnt = self.failed_conversion_cnt + 1

    def set_num_converted(self, num_converted_):
        self.num_converted = num_converted_

    def _create_categories(self, path_to_classes):

        classes = None
        s_class = namedtuple('Class', ['class_name', 'id'])

        with open(path_to_classes, 'r') as fp:
            classes = json.load(fp)
        categories = [
            self._create_single_category(s_class(item, classes[item]))
            for item in classes
        ]
        return categories

    def _make_id_generator(self):
        cur_id = 0
        while True:
            cur_id += 1
            yield cur_id

    def _create_skeleton(self):
        out_json = {
            'info':
                {
                    'description':
                        'This is {} dataset.'.format(self.dataset_name),
                    'url':
                        'https://superannotate.ai',
                    'version':
                        '1.0',
                    'year':
                        2020,
                    'contributor':
                        'Superannotate AI',
                    'date_created':
                        datetime.now().strftime("%d/%m/%Y")
                },
            'licenses':
                [
                    {
                        'url': 'https://superannotate.ai',
                        'id': 1,
                        'name': 'Superannotate AI'
                    }
                ],
            'images': [],
            'annotations': [],
            'categories': []
        }
        return out_json

    def _load_sa_jsons(self):
        if self.project_type == 'Pixel':
            jsons_gen = self.export_root.glob('*pixel.json')
        elif self.project_type == 'Vector':
            jsons_gen = self.export_root.glob('*objects.json')
        jsons = list(jsons_gen)
        self.set_num_converted(len(jsons))
        return jsons

    def _prepare_single_image_commons_pixel(self, id_, json_path):
        ImgCommons = namedtuple(
            'ImgCommons', [
                'image_info', 'sa_ann_json', 'ann_mask', 'sa_bluemask_rgb',
                'flat_mask'
            ]
        )
        rm_len = len('___pixel.json')

        sa_ann_json = json.load(open(json_path))
        sa_bluemask_path = str(json_path)[:-rm_len] + '___save.png'

        image_info = self.__make_image_info(json_path, id_, self.project_type)

        sa_bluemask_rgb = np.asarray(
            Image.open(sa_bluemask_path).convert('RGB'), dtype=np.uint32
        )

        ann_mask = np.zeros(
            (image_info['height'], image_info['width']), dtype=np.uint32
        )
        flat_mask = (sa_bluemask_rgb[:, :, 0] <<
                     16) | (sa_bluemask_rgb[:, :, 1] <<
                            8) | (sa_bluemask_rgb[:, :, 2])

        res = ImgCommons(
            image_info, sa_ann_json, ann_mask, sa_bluemask_rgb, flat_mask
        )

        return res

    def __make_image_info(self, json_path, id_, source_type):
        if source_type == 'Pixel':
            rm_len = len('___pixel.json')
        elif source_type == 'Vector':
            rm_len = len('___objects.json')

        image_path = str(json_path)[:-rm_len]

        img_width, img_height = Image.open(image_path).size
        image_info = {
            'id': id_,
            'file_name': Path(image_path).name,
            'height': img_height,
            'width': img_width,
            'license': 1
        }

        return image_info

    def _prepare_single_image_commons_vector(self, id_, json_path):
        ImgCommons = namedtuple('ImgCommons', ['image_info', 'sa_ann_json'])

        image_info = self.__make_image_info(json_path, id_, self.project_type)
        sa_ann_json = json.load(open(json_path))

        res = ImgCommons(image_info, sa_ann_json)

        return res

    def _prepare_single_image_commons(self, id_, json_path):
        res = None
        if self.project_type == 'Pixel':
            res = self._prepare_single_image_commons_pixel(id_, json_path)
        elif self.project_type == 'Vector':
            res = self._prepare_single_image_commons_vector(id_, json_path)
        return res

    def _create_sa_classes(self, json_path):
        json_data = json.load(open(json_path))
        classes_list = json_data["categories"]

        colors = self._generate_colors(len(classes_list))
        classes = []
        for c, data in enumerate(classes_list):
            classes_dict = {
                'name': data["name"],
                'id': data["id"],
                'color': colors[c],
                'attribute_groups': []
            }
            classes.append(classes_dict)

        return classes

    def _generate_colors(self, number):
        colors = []
        for _ in range(number):
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            colors.append(hexcolor)
        return colors



import json

from .yolo_converter import YOLOConverter
from .yolo_to_sa_vector import yolo_object_detection_to_sa_vector

from ....common import dump_output


class YoloObjectDetectionStrategy(YOLOConverter):
    name = "ObjectDetection converter"

    def __init__(self, args):
        super().__init__(args)
        self.__setup_conversion_algorithm()

    def __setup_conversion_algorithm(self):
        if self.direction == "from":
            if self.project_type == "Vector":
                if self.task == "object_detection":
                    self.converion_algorithm = yolo_object_detection_to_sa_vector

    def to_sa_format(self):
        sa_jsons, sa_classes = self.converion_algorithm(self.export_root)
        dump_output(self.output_dir, self.platform, sa_classes, sa_jsons)



import json
import os
import cv2

from .voc_converter import VocConverter
from .voc_to_sa_pixel import voc_instance_segmentation_to_sa_pixel
from .voc_to_sa_vector import voc_object_detection_to_sa_vector, voc_instance_segmentation_to_sa_vector

from ....common import dump_output


class VocObjectDetectionStrategy(VocConverter):
    name = "ObjectDetection converter"

    def __init__(self, args):
        super().__init__(args)
        self.__setup_conversion_algorithm()

    def __setup_conversion_algorithm(self):
        if self.direction == "to":
            raise NotImplementedError("Doesn't support yet")
        else:
            if self.project_type == "Vector":
                if self.task == "object_detection":
                    self.conversion_algorithm = voc_object_detection_to_sa_vector
                elif self.task == "instance_segmentation":
                    self.conversion_algorithm = voc_instance_segmentation_to_sa_vector
            elif self.project_type == "Pixel":
                if self.task == "instance_segmentation":
                    self.conversion_algorithm = voc_instance_segmentation_to_sa_pixel

    def __str__(self):
        return '{} object'.format(self.name)

    def to_sa_format(self):
        sa_classes, sa_jsons, sa_masks = self.conversion_algorithm(
            self.export_root
        )
        dump_output(self.output_dir, self.platform, sa_classes, sa_jsons)

        if self.project_type == 'Pixel':
            all_files = self.output_dir.glob('*.png')
            for file in all_files:
                if '___save.png' not in str(file.name):
                    (self.output_dir / file.name).unlink()

            for sa_mask_name, sa_mask_value in sa_masks.items():
                sa_mask_value = sa_mask_value[:, :, ::-1]
                cv2.imwrite(str(self.output_dir / sa_mask_name), sa_mask_value)



import json
import numpy as np

from pathlib import Path


def _create_classes(classes):
    sa_classes_loader = []
    for key, value in classes.items():
        color = np.random.choice(range(256), size=3)
        hexcolor = "#%02x%02x%02x" % tuple(color)
        sa_classes = {'name': key, 'color': hexcolor, 'attribute_groups': []}
        attribute_groups = []
        for attr_group_key, attr_group in value['attribute_group'].items():
            attr_loader = {
                'name': attr_group_key,
                'is_multiselect': attr_group['is_multiselect'],
                'attributes': []
            }
            for attr in set(attr_group['attributes']):
                attr_loader['attributes'].append({'name': attr})
            if attr_loader:
                attribute_groups.append(attr_loader)

        sa_classes['attribute_groups'] = attribute_groups

        sa_classes_loader.append(sa_classes)
    return sa_classes_loader


def dataloop_object_detection_to_sa_vector(input_dir):
    classes = {}
    sa_jsons = {}
    json_data = Path(input_dir).glob('*.json')
    sa_classes_labels = ['point', 'box', 'ellipse', 'segment', 'binary']
    for json_file in json_data:
        sa_loader = []
        dl_data = json.load(open(json_file))

        for ann in dl_data['annotations']:
            if ann['type'] in sa_classes_labels:
                if ann['label'] not in classes.keys():
                    classes[ann['label']] = {'attribute_group': {}}
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'] = {}
                    classes[ann['label']]['attribute_group'][
                        'converted_attributes']['is_multiselect'] = 1
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'][
                               'attributes'] = ann['attributes']
                else:
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'][
                               'attributes'] += ann['attributes']

            attributes = []
            for attribute in ann['attributes']:
                attr = {'name': attribute, 'groupName': 'converted_attributes'}
                attributes.append(attr)

            if ann['type'] == 'box':
                sa_bbox = {
                    'type': 'bbox',
                    'points':
                        {
                            'x1': ann['coordinates'][0]['x'],
                            'y1': ann['coordinates'][0]['y'],
                            'x2': ann['coordinates'][1]['x'],
                            'y2': ann['coordinates'][1]['y']
                        },
                    'className': ann['label'],
                    'attributes': attributes,
                    'probability': 100,
                    'locked': False,
                    'visible': True,
                    'groupId': 0
                }
                sa_loader.append(sa_bbox)
            elif ann['type'] == 'note':
                sa_comment = {
                    'type': 'comment',
                    'x': ann['coordinates']['box'][0]['x'],
                    'y': ann['coordinates']['box'][0]['y'],
                    'comments': []
                }
                for note in ann['coordinates']['note']['messages']:
                    sa_comment['comments'].append(
                        {
                            'text': note['body'],
                            'id': note['creator']
                        }
                    )
                sa_loader.append(sa_comment)
            elif ann['type'] == 'class':
                sa_tags = {'type': 'tag', 'name': ann['label']}
                sa_loader.append(sa_tags)

        file_name = dl_data['filename'][1:] + '___objects.json'
        sa_jsons[file_name] = sa_loader

    classes = _create_classes(classes)
    return sa_jsons, classes


def dataloop_instance_segmentation_to_sa_vector(input_dir):
    classes = {}
    sa_jsons = {}
    json_data = Path(input_dir).glob('*.json')
    sa_classes_labels = ['point', 'box', 'ellipse', 'segment', 'binary']
    for json_file in json_data:
        sa_loader = []
        dl_data = json.load(open(json_file))

        for ann in dl_data['annotations']:
            if ann['type'] in sa_classes_labels:
                if ann['label'] not in classes.keys():
                    classes[ann['label']] = {'attribute_group': {}}
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'] = {}
                    classes[ann['label']]['attribute_group'][
                        'converted_attributes']['is_multiselect'] = 1
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'][
                               'attributes'] = ann['attributes']
                else:
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'][
                               'attributes'] += ann['attributes']

            attributes = []
            for attribute in ann['attributes']:
                attr = {'name': attribute, 'groupName': 'converted_attributes'}
                attributes.append(attr)

            if ann['type'] == 'segment' and len(ann['coordinates']) == 1:
                sa_polygon = {
                    'type': 'polygon',
                    'points': [],
                    'className': ann['label'],
                    'attributes': attributes,
                    'probability': 100,
                    'locked': False,
                    'visible': True,
                    'groupId': 0,
                }
                for sub_list in ann['coordinates']:
                    for sub_dict in sub_list:
                        sa_polygon['points'].append(sub_dict['x'])
                        sa_polygon['points'].append(sub_dict['y'])

                sa_loader.append(sa_polygon)
            elif ann['type'] == 'note':
                sa_comment = {
                    'type': 'comment',
                    'x': ann['coordinates']['box'][0]['x'],
                    'y': ann['coordinates']['box'][0]['y'],
                    'comments': []
                }
                for note in ann['coordinates']['note']['messages']:
                    sa_comment['comments'].append(
                        {
                            'text': note['body'],
                            'id': note['creator']
                        }
                    )
                sa_loader.append(sa_comment)
            elif ann['type'] == 'class':
                sa_tags = {'type': 'tag', 'name': ann['label']}
                sa_loader.append(sa_tags)

        file_name = dl_data['filename'][1:] + '___objects.json'
        sa_jsons[file_name] = sa_loader

    classes = _create_classes(classes)
    return sa_jsons, classes


def dataloop_to_sa(input_dir):
    classes = {}
    sa_jsons = {}
    json_data = Path(input_dir).glob('*.json')
    sa_classes_labels = ['point', 'box', 'ellipse', 'segment', 'binary']
    for json_file in json_data:
        sa_loader = []
        dl_data = json.load(open(json_file))

        for ann in dl_data['annotations']:
            if ann['type'] in sa_classes_labels:
                if ann['label'] not in classes.keys():
                    classes[ann['label']] = {'attribute_group': {}}
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'] = {}
                    classes[ann['label']]['attribute_group'][
                        'converted_attributes']['is_multiselect'] = 1
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'][
                               'attributes'] = ann['attributes']
                else:
                    classes[ann['label']
                           ]['attribute_group']['converted_attributes'][
                               'attributes'] += ann['attributes']

            attributes = []
            for attribute in ann['attributes']:
                attr = {'name': attribute, 'groupName': 'converted_attributes'}
                attributes.append(attr)

            if ann['type'] == 'segment' and len(ann['coordinates']) == 1:
                sa_polygon = {
                    'type': 'polygon',
                    'points': [],
                    'className': ann['label'],
                    'attributes': attributes,
                    'probability': 100,
                    'locked': False,
                    'visible': True,
                    'groupId': 0,
                }
                for sub_list in ann['coordinates']:
                    for sub_dict in sub_list:
                        sa_polygon['points'].append(sub_dict['x'])
                        sa_polygon['points'].append(sub_dict['y'])
                sa_loader.append(sa_polygon)
            elif ann['type'] == 'box':
                sa_bbox = {
                    'type': 'bbox',
                    'points':
                        {
                            'x1': ann['coordinates'][0]['x'],
                            'y1': ann['coordinates'][0]['y'],
                            'x2': ann['coordinates'][1]['x'],
                            'y2': ann['coordinates'][1]['y']
                        },
                    'className': ann['label'],
                    'attributes': attributes,
                    'probability': 100,
                    'locked': False,
                    'visible': True,
                    'groupId': 0
                }
                sa_loader.append(sa_bbox)
            elif ann['type'] == 'ellipse':
                sa_ellipse = {
                    'type': 'ellipse',
                    'className': ann['label'],
                    'probability': 100,
                    'cx': ann['coordinates']['center']['x'],
                    'cy': ann['coordinates']['center']['y'],
                    'rx': ann['coordinates']['rx'],
                    'ry': ann['coordinates']['ry'],
                    'angle': ann['coordinates']['angle'],
                    'groupId': 0,
                    'pointLabels': {},
                    'attributes': attributes,
                    'locked': False,
                    'visible': True,
                    'attributes': []
                }
                sa_loader.append(sa_ellipse)
            elif ann['type'] == 'point':
                sa_point = {
                    'type': 'point',
                    'className': ann['label'],
                    'probability': 100,
                    'x': ann['coordinates']['x'],
                    'y': ann['coordinates']['y'],
                    'groupId': 0,
                    'pointLabels': {},
                    'attributes': attributes,
                    'locked': False,
                    'visible': True,
                    'attributes': []
                }
                sa_loader.append(sa_point)
            elif ann['type'] == 'note':
                sa_comment = {
                    'type': 'comment',
                    'x': ann['coordinates']['box'][0]['x'],
                    'y': ann['coordinates']['box'][0]['y'],
                    'comments': []
                }
                for note in ann['coordinates']['note']['messages']:
                    sa_comment['comments'].append(
                        {
                            'text': note['body'],
                            'id': note['creator']
                        }
                    )
                sa_loader.append(sa_comment)
            elif ann['type'] == 'class':
                sa_tags = {'type': 'tag', 'name': ann['label']}
                sa_loader.append(sa_tags)

        file_name = dl_data['filename'][1:] + '___objects.json'
        sa_jsons[file_name] = sa_loader

    classes = _create_classes(classes)
    return sa_jsons, classes



import numpy as np
import os
import json
import zlib
import base64
import cv2


# Converts bitmaps to polygon
def _base64_to_polygon(bitmap):
    z = zlib.decompress(base64.b64decode(bitmap))
    n = np.frombuffer(z, np.uint8)
    mask = cv2.imdecode(n, cv2.IMREAD_UNCHANGED)[:, :, 3].astype(bool)
    contours, _ = cv2.findContours(
        mask.astype(np.uint8), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    segmentation = []

    for contour in contours:
        contour = contour.flatten().tolist()
        if len(contour) > 4:
            segmentation.append(contour)
        if len(segmentation) == 0:
            continue
    return segmentation


def supervisely_to_sa(json_files, class_id_map):
    sa_jsons = {}
    for json_file in json_files:
        file_name = os.path.splitext(os.path.basename(json_file)
                                    )[0] + '___objects.json'

        json_data = json.load(open(json_file))
        sa_loader = []

        for obj in json_data['objects']:
            if 'classTitle' in obj and obj['classTitle'] in class_id_map.keys():
                attributes = []
                if 'tags' in obj.keys():
                    for tag in obj['tags']:
                        group_name = class_id_map[obj['classTitle']
                                                 ]['attr_group']['group_name']
                        attr_name = tag['name']
                        attributes.append(
                            {
                                'name': attr_name,
                                'groupName': group_name
                            }
                        )

                    sa_obj = {
                        'type': '',
                        'points': [],
                        'className': obj['classTitle'],
                        'pointLabels': {},
                        'attributes': attributes,
                        'probability': 100,
                        'locked': False,
                        'visible': True,
                        'groupId': 0
                    }

                    if obj['geometryType'] == 'point':
                        del sa_obj['points']
                        sa_obj['type'] = 'point'
                        sa_obj['x'] = obj['points']['exterior'][0][0]
                        sa_obj['y'] = obj['points']['exterior'][0][1]

                    elif obj['geometryType'] == 'line':
                        sa_obj['type'] = 'polyline'
                        sa_obj['points'] = [
                            item for el in obj['points']['exterior']
                            for item in el
                        ]

                    elif obj['geometryType'] == 'rectangle':
                        sa_obj['type'] = 'bbox'
                        sa_obj['points'] = {
                            'x1': obj['points']['exterior'][0][0],
                            'y1': obj['points']['exterior'][0][1],
                            'x2': obj['points']['exterior'][1][0],
                            'y2': obj['points']['exterior'][1][1]
                        }

                    elif obj['geometryType'] == 'polygon':
                        sa_obj['type'] = 'polygon'
                        sa_obj['points'] = [
                            item for el in obj['points']['exterior']
                            for item in el
                        ]

                    elif obj['geometryType'] == 'cuboid':
                        sa_obj['type'] = 'cuboid'
                        sa_obj['points'] = {
                            'f1':
                                {
                                    'x': obj['points'][0][0],
                                    'y': obj['points'][0][1]
                                },
                            'f2':
                                {
                                    'x': obj['points'][2][0],
                                    'y': obj['points'][2][1],
                                },
                            'r1':
                                {
                                    'x': obj['points'][4][0],
                                    'y': obj['points'][4][1]
                                },
                            'r2':
                                {
                                    'x': obj['points'][5][0],
                                    'y': obj['points'][6][1],
                                }
                        }
                    elif obj['geometryType'] == 'bitmap':
                        for ppoints in _base64_to_polygon(
                            obj['bitmap']['data']
                        ):
                            sa_ppoints = [
                                x + obj['bitmap']['origin'][0] if i %
                                2 == 0 else x + obj['bitmap']['origin'][1]
                                for i, x in enumerate(ppoints)
                            ]
                            sa_obj['type'] = 'polygon'
                            sa_obj['points'] = sa_ppoints

                    sa_loader.append(sa_obj)
        sa_jsons[file_name] = sa_loader
    return sa_jsons


def supervisely_instance_segmentation_to_sa_vector(json_files, class_id_map):
    sa_jsons = {}
    for json_file in json_files:
        file_name = os.path.splitext(os.path.basename(json_file)
                                    )[0] + '___objects.json'

        json_data = json.load(open(json_file))
        sa_loader = []

        for obj in json_data['objects']:
            if 'classTitle' in obj and obj['classTitle'] in class_id_map.keys():
                attributes = []
                if 'tags' in obj.keys():
                    for tag in obj['tags']:
                        group_name = class_id_map[obj['classTitle']
                                                 ]['attr_group']['group_name']
                        attr_name = tag['name']
                        attributes.append(
                            {
                                'name': attr_name,
                                'groupName': group_name
                            }
                        )

                    sa_obj = {
                        'type': '',
                        'points': [],
                        'className': obj['classTitle'],
                        'pointLabels': {},
                        'attributes': attributes,
                        'probability': 100,
                        'locked': False,
                        'visible': True,
                        'groupId': 0
                    }

                    if obj['geometryType'] == 'polygon':
                        sa_obj['type'] = 'polygon'
                        sa_obj['points'] = [
                            item for el in obj['points']['exterior']
                            for item in el
                        ]
                    sa_loader.append(sa_obj)
        sa_jsons[file_name] = sa_loader
    return sa_jsons


def supervisely_object_detection_to_sa_vector(json_files, class_id_map):
    sa_jsons = {}
    for json_file in json_files:
        file_name = os.path.splitext(os.path.basename(json_file)
                                    )[0] + '___objects.json'

        json_data = json.load(open(json_file))
        sa_loader = []

        for obj in json_data['objects']:
            if 'classTitle' in obj and obj['classTitle'] in class_id_map.keys():
                attributes = []
                if 'tags' in obj.keys():
                    for tag in obj['tags']:
                        group_name = class_id_map[obj['classTitle']
                                                 ]['attr_group']['group_name']
                        attr_name = tag['name']
                        attributes.append(
                            {
                                'name': attr_name,
                                'groupName': group_name
                            }
                        )

                    sa_obj = {
                        'type': '',
                        'points': [],
                        'className': obj['classTitle'],
                        'pointLabels': {},
                        'attributes': attributes,
                        'probability': 100,
                        'locked': False,
                        'visible': True,
                        'groupId': 0
                    }

                    if obj['geometryType'] == 'rectangle':
                        sa_obj['type'] = 'bbox'
                        sa_obj['points'] = {
                            'x1': obj['points']['exterior'][0][0],
                            'y1': obj['points']['exterior'][0][1],
                            'x2': obj['points']['exterior'][1][0],
                            'y2': obj['points']['exterior'][1][1]
                        }
                    elif obj['geometryType'] == 'polygon':
                        sa_obj['type'] = 'bbox'
                        segment = [
                            item for el in obj['points']['exterior']
                            for item in el
                        ]
                        sa_obj['points'] = {
                            'x1': min(segment[::2]),
                            'y1': min(segment[1::2]),
                            'x2': max(segment[::2]),
                            'y2': max(segment[1::2])
                        }
                    sa_loader.append(sa_obj)
        sa_jsons[file_name] = sa_loader
    return sa_jsons


def supervisely_keypoint_detection_to_sa_vector(
    json_files, class_id_map, meta_json
):
    classes_skeleton = {}
    for class_ in meta_json['classes']:
        if class_['shape'] != 'graph':
            continue

        classes_skeleton[class_['title']] = {'edges': [], 'nodes': {}}
        nodes = class_['geometry_config']['nodes']
        for node, value in nodes.items():
            classes_skeleton[class_['title']]['nodes'][node] = value['label']

        edges = class_['geometry_config']['edges']
        for edge in edges:
            classes_skeleton[class_['title']]['edges'].append(
                (edge['src'], edge['dst'])
            )

    sa_jsons = {}
    for json_file in json_files:
        file_name = os.path.splitext(os.path.basename(json_file)
                                    )[0] + '___objects.json'

        json_data = json.load(open(json_file))
        sa_loader = []

        for obj in json_data['objects']:
            if 'classTitle' in obj and obj['classTitle'] in class_id_map.keys():
                attributes = []
                if 'tags' in obj.keys():
                    for tag in obj['tags']:
                        group_name = class_id_map[obj['classTitle']
                                                 ]['attr_group']['group_name']
                        attr_name = tag['name']
                        attributes.append(
                            {
                                'name': attr_name,
                                'groupName': group_name
                            }
                        )

                    sa_obj = {
                        'type': '',
                        'points': [],
                        'className': obj['classTitle'],
                        'pointLabels': {},
                        'attributes': attributes,
                        'probability': 100,
                        'locked': False,
                        'visible': True,
                        'connections': []
                    }

                    if obj['geometryType'] == 'graph':
                        sa_obj['type'] = 'template'
                        good_nodes = []
                        nodes = obj['nodes']
                        index = 1
                        for node, value in nodes.items():
                            good_nodes.append(node)
                            sa_obj['points'].append(
                                {
                                    'id': index,
                                    'x': value['loc'][0],
                                    'y': value['loc'][1]
                                }
                            )
                            sa_obj['pointLabels'][index - 1] = classes_skeleton[
                                obj['classTitle']]['nodes'][node]
                            index += 1

                        index = 1
                        for edge in classes_skeleton[obj['classTitle']
                                                    ]['edges']:
                            if edge[0] not in good_nodes or edge[
                                1] not in good_nodes:
                                continue

                            sa_obj['connections'].append(
                                {
                                    'id': index,
                                    'from': good_nodes.index(edge[0]) + 1,
                                    'to': good_nodes.index(edge[1]) + 1,
                                }
                            )
                            index += 1

                    sa_loader.append(sa_obj)
        sa_jsons[file_name] = sa_loader
    return sa_jsons



import cv2
import os
from glob import glob

from .sagemaker_converter import SageMakerConverter
from .sagemaker_to_sa_vector import sagemaker_object_detection_to_sa_vector
from .sagemaker_to_sa_pixel import sagemaker_instance_segmentation_to_sa_pixel

from ....common import dump_output


class SageMakerObjectDetectionStrategy(SageMakerConverter):
    name = "ObjectDetection converter"

    def __init__(self, args):
        super().__init__(args)
        self.__setup_conversion_algorithm()

    def __setup_conversion_algorithm(self):
        if self.direction == "from":
            if self.project_type == "Vector":
                if self.task == "object_detection":
                    self.conversion_algorithm = sagemaker_object_detection_to_sa_vector
            elif self.project_type == "Pixel":
                if self.task == "instance_segmentation":
                    self.conversion_algorithm = sagemaker_instance_segmentation_to_sa_pixel

    def __str__(self):
        return '{} object'.format(self.name)

    def to_sa_format(self):
        if self.conversion_algorithm.__name__ == 'sagemaker_object_detection_to_sa_vector':
            sa_jsons, sa_classes, sa_masks = self.conversion_algorithm(
                self.export_root, self.dataset_name
            )
        else:
            sa_jsons, sa_classes, sa_masks = self.conversion_algorithm(
                self.export_root
            )

        old_masks = self.output_dir.glob('*.png')
        for mask in old_masks:
            mask.unlink()
        if self.project_type == 'Pixel':
            for name, mask in sa_masks.items():
                cv2.imwrite(str(self.output_dir / name), mask)

        dump_output(self.output_dir, self.platform, sa_classes, sa_jsons)



"""
Module which will run converters and convert from other
annotation formats to superannotate annotation format
"""
import logging
import shutil
from pathlib import Path

from .converters.converters import Converter

logger = logging.getLogger("superannotate-python-sdk")


def _load_files(path_to_imgs, ptype, extensions):
    images = []
    for extension in extensions:
        rec_search = str(Path('**') / ('*.' + extension))
        images_gen = Path(path_to_imgs).glob(rec_search)
        images.extend(list(images_gen))

    if not images:
        logger.warning("Images doesn't exist")

    if ptype == 'Pixel':
        rec_search = str(Path('**') / '*.png')
        masks_gen = Path(path_to_imgs).glob(rec_search)
        masks = list(masks_gen)
    else:
        masks = []

    return images, masks


def _move_files(imgs, masks, output_dir, platforom):
    if platforom == "Desktop":
        output_path = output_dir / "images"
        output_path.mkdir(parents=True)
    else:
        (output_dir / 'classes').mkdir(parents=True, exist_ok=True)
        output_path = output_dir

    for im in imgs:
        shutil.copy(im, output_path / Path(im).name)

    for mask in masks:
        shutil.copy(mask, output_path / Path(mask).name)


def import_to_sa(args):
    """
    :param args: All arguments that will be used during convertion.
    :type args: Namespace
    """

    images, masks = _load_files(
        args.input_dir / args.images_root, args.project_type,
        args.images_extensions
    )
    _move_files(images, masks, args.output_dir, args.platform)

    args.__dict__.update({'direction': 'from', 'export_root': args.input_dir})
    converter = Converter(args)

    converter.convert_to_sa(args.platform)

    logger.info('Conversion completed successfully')



import os
import json
import numpy as np


def _create_classes(classes_map):
    classes_loader = []
    for key, value in classes_map.items():
        color = np.random.choice(range(256), size=3)
        hexcolor = "#%02x%02x%02x" % tuple(color)
        sa_classes = {'name': value, 'color': hexcolor, 'attribute_groups': []}
        classes_loader.append(sa_classes)
    return classes_loader


def sagemaker_object_detection_to_sa_vector(data_path, main_key):
    sa_jsons = {}
    dataset_manifest = []
    try:
        img_map_file = open(data_path / 'output.manifest')
    except Exception as e:
        raise Exception("'output.manifest' file doesn't exist")

    for line in img_map_file:
        dataset_manifest.append(json.loads(line))

    json_list = data_path.glob('*.json')
    classes_ids = {}
    for json_file in json_list:
        data_json = json.load(open(json_file))
        for img in data_json:
            if 'consolidatedAnnotation' not in img.keys():
                print('Wrong json files')
                raise Exception

            manifest = dataset_manifest[int(img['datasetObjectId'])]
            file_name = os.path.basename(
                manifest['source-ref']
            ) + '___objects.json'

            classes = img['consolidatedAnnotation']['content'][
                main_key + '-metadata']['class-map']
            for key, value in classes.items():
                if key not in classes_ids.keys():
                    classes_ids[key] = value

            annotations = img['consolidatedAnnotation']['content'][main_key][
                'annotations']
            sa_loader = []
            for annotation in annotations:
                points = {
                    'x1': annotation['left'],
                    'y1': annotation['top'],
                    'x2': annotation['left'] + annotation['width'],
                    'y2': annotation['top'] + annotation['height']
                }
                sa_obj = {
                    'type': 'bbox',
                    'points': points,
                    'className': classes[str(annotation['class_id'])],
                    'attributes': [],
                    'probability': 100,
                    'locked': False,
                    'visible': True,
                    'groupId': 0
                }
                sa_loader.append(sa_obj.copy())
            sa_jsons[file_name] = sa_loader

    sa_classes = _create_classes(classes_ids)
    return sa_jsons, sa_classes, None



import logging

from . import common

logger = logging.getLogger("superannotate-python-sdk")


class SABaseException(Exception):
    def __init__(
        self,
        status_code='Unknown Status Code',
        message='Unknown Error Message'
    ):
        self.status_code = status_code
        self.message = message
        super().__init__(status_code, message)


class SAImageSizeTooLarge(SABaseException):
    def __init__(self, file_size, file_name=""):
        super().__init__(
            0, "Image " + file_name + " size " + str(file_size // 1024**2) +
            " MB is larger than " + str(common.MAX_IMAGE_SIZE // 1024**2) +
            " MB limit."
        )


class SAExistingProjectNameException(SABaseException):
    pass


class SANonExistingProjectNameException(SABaseException):
    pass


class SAExistingAnnotationClassNameException(SABaseException):
    pass


class SANonExistingAnnotationClassNameException(SABaseException):
    pass


class SAExistingExportNameException(SABaseException):
    pass


class SANonExistingExportNameException(SABaseException):
    pass



import os
import cv2
import json
from tqdm import tqdm
import numpy as np
import xml.etree.ElementTree as ET

from ....common import hex_to_rgb, blue_color_generator


# Generates polygons for each instance
def _generate_polygons(object_mask_path, class_mask_path):
    segmentation = []

    object_mask = cv2.imread(str(object_mask_path), cv2.IMREAD_GRAYSCALE)
    class_mask = cv2.imread(str(class_mask_path), cv2.IMREAD_GRAYSCALE)

    object_unique_colors = np.unique(object_mask)

    num_colors = len([i for i in object_unique_colors if i not in (0, 220)])
    bluemask_colors = blue_color_generator(num_colors)
    H, W = object_mask.shape
    sa_mask = np.zeros((H, W, 4))
    i = 0
    for unique_color in object_unique_colors:
        if unique_color in (0, 220):
            continue

        # class_color = class_mask[object_mask == unique_color][0]
        mask = np.zeros_like(object_mask)
        mask[object_mask == unique_color] = 255
        contours, _ = cv2.findContours(
            mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )

        segment = []
        for contour in contours:
            contour = contour.flatten().tolist()
            segment += contour
        if len(contour) > 4:
            segmentation.append(segment)
        if len(segmentation) == 0:
            continue

        sa_mask[object_mask == unique_color
               ] = [255] + list(hex_to_rgb(bluemask_colors[i]))
        i += 1
    return segmentation, sa_mask, bluemask_colors


def _iou(bbox1, bbox2):
    xmin1, ymin1, xmax1, ymax1 = bbox1
    xmin2, ymin2, xmax2, ymax2 = bbox2

    x = max(0, min(xmax1, xmax2) - max(xmin1, xmin2))
    y = max(0, min(ymax1, ymax2) - max(ymin1, ymin2))
    return x * y / float(
        (xmax1 - xmin1) * (ymax1 - ymin1) + (xmax2 - xmin2) *
        (ymax2 - ymin2) - x * y
    )


def _generate_instances(polygon_instances, voc_instances, bluemask_colors):
    instances = []
    i = 0
    # for polygon, color_id in polygon_instances:
    for polygon in polygon_instances:
        ious = []
        bbox_poly = [
            min(polygon[::2]),
            min(polygon[1::2]),
            max(polygon[::2]),
            max(polygon[1::2])
        ]
        for class_name, bbox in voc_instances:
            ious.append(_iou(bbox_poly, bbox))
        ind = np.argmax(ious)
        instances.append(
            {
                "className": voc_instances[ind][0],
                # "classId": color_id,
                "polygon": polygon,
                "bbox": voc_instances[ind][1],
                "blue_color": bluemask_colors[i]
            }
        )
        i += 1
    return instances


def _get_voc_instances_from_xml(file_path):
    with open(os.path.splitext(file_path)[0] + ".xml") as f:
        tree = ET.parse(f)
    instances = tree.findall('object')
    voc_instances = []
    for instance in instances:
        class_name = instance.find("name").text
        bbox = instance.find("bndbox")
        bbox = [
            float(bbox.find(x).text) for x in ["xmin", "ymin", "xmax", "ymax"]
        ]
        voc_instances.append((class_name, bbox))
    return voc_instances


def _create_classes(classes):
    sa_classes = []
    # for class_, id_ in classes.items():
    for class_ in set(classes):
        color = np.random.choice(range(256), size=3)
        hexcolor = "#%02x%02x%02x" % tuple(color)
        sa_class = {
            # "id": id_,
            "name": class_,
            "color": hexcolor,
            "attribute_groups": []
        }
        sa_classes.append(sa_class)
    return sa_classes


def voc_instance_segmentation_to_sa_pixel(voc_root):
    classes = []
    object_masks_dir = voc_root / 'SegmentationObject'
    class_masks_dir = voc_root / 'SegmentationClass'
    annotation_dir = voc_root / "Annotations"

    sa_jsons = {}
    sa_masks = {}
    for filename in object_masks_dir.glob('*'):
        polygon_instances, sa_mask, bluemask_colors = _generate_polygons(
            object_masks_dir / filename.name,
            class_masks_dir / filename.name,
        )
        voc_instances = _get_voc_instances_from_xml(
            annotation_dir / filename.name
        )
        maped_instances = _generate_instances(
            polygon_instances, voc_instances, bluemask_colors
        )

        sa_loader = []
        for instance in maped_instances:
            sa_polygon = {
                'className': instance["className"],
                'parts': [{
                    "color": instance["blue_color"]
                }],
                'attributes': [],
                'probability': 100,
                'locked': False,
                'visible': True,
            }
            sa_loader.append(sa_polygon)
            classes.append(instance['className'])

        sa_file_name = os.path.splitext(filename.name)[0] + ".jpg___pixel.json"
        sa_jsons[sa_file_name] = sa_loader

        sa_mask_name = os.path.splitext(filename.name)[0] + ".jpg___save.png"
        sa_masks[sa_mask_name] = sa_mask

    classes = _create_classes(classes)
    return (classes, sa_jsons, sa_masks)



import json
import logging
from pathlib import Path

import cv2
import numpy as np
from tqdm import tqdm

from ....common import blue_color_generator, hex_to_rgb, id2rgb
from ....pycocotools_sa.coco import COCO
from ....pycocotools_sa import mask as maskUtils

logger = logging.getLogger("superannotate-python-sdk")


def _rle_to_polygon(coco_json, annotation):
    coco = COCO(coco_json)
    binary_mask = coco.annToMask(annotation)
    contours, _ = cv2.findContours(
        binary_mask.astype(np.uint8), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    segmentation = []

    for contour in contours:
        contour = contour.flatten().tolist()
        if len(contour) > 4:
            segmentation.append(contour)
        if len(segmentation) == 0:
            continue
    return segmentation


def coco_panoptic_segmentation_to_sa_pixel(coco_path, images_path):
    coco_json = json.load(open(coco_path))
    hex_colors = blue_color_generator(len(coco_json["categories"]))
    annotate_list = coco_json["annotations"]

    cat_id_to_cat = {}
    for cat in coco_json['categories']:
        cat_id_to_cat[cat['id']] = cat['name']

    sa_jsons = {}
    for annotate in tqdm(annotate_list, "Converting annotations"):
        annot_name = Path(annotate["file_name"]).stem
        img_cv = cv2.imread(str(images_path / (annot_name + ".png")))
        if img_cv is None:
            logger.warning(
                "'{}' file dosen't exist!".format(
                    images_path / (annot_name + ".png")
                )
            )
            continue

        img = cv2.cvtColor(img_cv, cv2.COLOR_BGR2RGB)
        H, W, C = img.shape
        img = img.reshape((H * W, C))
        segments = annotate["segments_info"]
        hex_colors = blue_color_generator(len(segments))

        out_json = []
        for i, seg in enumerate(segments):
            img[np.all(img == id2rgb(seg["id"]),
                       axis=1)] = hex_to_rgb(hex_colors[i])
            dd = {
                "classId": seg["category_id"],
                'className': cat_id_to_cat[seg["category_id"]],
                "probability": 100,
                "visible": True,
                "parts": [{
                    "color": hex_colors[i]
                }],
                "attributes": [],
                "attributeNames": [],
                "imageId": annotate["image_id"]
            }
            out_json.append(dd)

        img = cv2.cvtColor(img.reshape((H, W, C)), cv2.COLOR_RGB2BGR)
        cv2.imwrite(str(images_path / (annot_name + ".jpg___save.png")), img)

        file_name = annot_name + ".jpg___pixel.json"
        sa_jsons[file_name] = out_json
        (images_path / (annot_name + ".png")).unlink()
    return sa_jsons


def coco_instance_segmentation_to_sa_pixel(coco_path, images_path):
    coco_json = json.load(open(coco_path))
    cat_id_to_cat = {}
    for cat in coco_json['categories']:
        cat_id_to_cat[cat['id']] = cat

    images_dict = {}
    for img in coco_json['images']:
        images_dict[img['id']] = {
            'shape': (img['height'], img['width'], 4),
            'file_name': img['file_name'],
            'annotations': []
        }

    for annot in coco_json['annotations']:
        if annot['image_id'] not in images_dict:
            continue
        images_dict[annot['image_id']]['annotations'].append(annot)

    sa_jsons = {}
    for id_, annotations in images_dict.items():
        file_name = annotations['file_name'] + '___pixel.json'
        hexcolors = blue_color_generator(len(annotations['annotations']))
        mask = np.zeros(annotations['shape'])
        H, W, _ = mask.shape

        sa_loader = []
        for i, annot in enumerate(annotations['annotations']):
            hexcolor = hexcolors[i]
            color = hex_to_rgb(hexcolor)
            if isinstance(annot['segmentation'], dict):
                if isinstance(annot['segmentation']['counts'], list):
                    annot['segmentation'] = _rle_to_polygon(coco_path, annot)
                    for segment in annot['segmentation']:
                        bitmask = np.zeros((H, W)).astype(np.uint8)
                        pts = np.array(
                            [
                                segment[2 * i:2 * (i + 1)]
                                for i in range(len(segment) // 2)
                            ],
                            dtype=np.int32
                        )

                        cv2.fillPoly(bitmask, [pts], 1)
                        mask[bitmask == 1] = list(color)[::-1] + [255]
                else:
                    mask[maskUtils.decode(annot['segmentation']) == 1
                        ] = list(color)[::-1] + [255]
            else:
                for segment in annot['segmentation']:
                    bitmask = np.zeros((H, W)).astype(np.uint8)
                    pts = np.array(
                        [
                            segment[2 * i:2 * (i + 1)]
                            for i in range(len(segment) // 2)
                        ],
                        dtype=np.int32
                    )
                    cv2.fillPoly(bitmask, [pts], 1)
                    mask[bitmask == 1] = list(color)[::-1] + [255]

            sa_obj = {
                "className": cat_id_to_cat[annot['category_id']]['name'],
                'probability': 100,
                'visible': True,
                'parts': [{
                    'color': hexcolor
                }],
                "attributes": [],
                "attributeNames": [],
            }
            sa_loader.append(sa_obj.copy())
        sa_jsons[file_name] = sa_loader
        cv2.imwrite(
            str(images_path / (annotations['file_name'] + '___save.png')), mask
        )

    return sa_jsons



from pathlib import Path

import numpy as np
import pydicom
from PIL import Image


def dicom_to_rgb_sequence(
    input_dicom_file, output_dir, output_image_quality="original"
):
    """Converts DICOM file to RGB image sequence.
    Output file format is <input_dicom_file_name_woextension>_<frame_number>.jpg

    :param input_dicom_file: path to DICOM file
    :type input_dicom_file: str or Pathlike
    :param output_dir: path to output directory
    :type output_dir: str or Pathlike
    :param output_image_quality: output quality "original" or "compressed"
    :type output_image_quality: str

    :return: paths to output images
    :rtype: list of strs

    """
    input_dicom_file = Path(input_dicom_file)
    ds = pydicom.dcmread(str(input_dicom_file))
    # array = np.frombuffer(ds[0x43, 0x1029].value, np.uint8)
    # # interp = ds.PhotometricInterpretation
    # np.set_printoptions(threshold=10000000)
    # print(array)

    arr = ds.pixel_array
    if "NumberOfFrames" in ds:
        number_of_frames = ds.NumberOfFrames
    else:
        number_of_frames = 1
        arr = arr[np.newaxis, :]
        if arr.dtype != np.uint8:
            arr = (arr - arr.min()) / arr.max() * 255
            arr = arr.astype(np.uint8)
    output_dir = Path(output_dir)
    output_paths = []
    for i in range(number_of_frames):
        image = Image.fromarray(arr[i])
        image = image.convert("RGB")
        path = output_dir / (input_dicom_file.stem + f"_{i:05}.jpg")
        image.save(
            path,
            subsampling=0 if output_image_quality == "original" else 2,
            quality=100 if output_image_quality == "original" else 60
        )
        output_paths.append(str(path))

    return output_paths



import os
import pandas as pd
import numpy as np
import cv2


def _create_classes(classes):
    classes_loader = []
    for class_ in set(classes):
        color = np.random.choice(range(256), size=3)
        hexcolor = "#%02x%02x%02x" % tuple(color)
        sa_classes = {'name': class_, 'color': hexcolor, 'attribute_groups': []}
        classes_loader.append(sa_classes)
    return classes_loader


def googlecloud_object_detection_to_sa_vector(path):
    df = pd.read_csv(path, header=None)
    dir_name = path.parent

    sa_jsons = {}
    classes = []
    for _, row in df.iterrows():
        classes.append(row[2])

        file_name = row[1].split('/')[-1]
        img = cv2.imread(str(dir_name / file_name))
        H, W, _ = img.shape
        sa_file_name = os.path.basename(file_name) + '___objects.json'
        xmin = row[3] * W
        xmax = row[5] * W
        ymin = row[4] * H
        ymax = row[8] * H

        sa_obj = {
            'type': 'bbox',
            'points': {
                'x1': xmin,
                'y1': ymin,
                'x2': xmax,
                'y2': ymax
            },
            'className': row[2],
            'attributes': [],
            'probability': 100,
            'locked': False,
            'visible': True,
            'groupId': 0
        }

        if sa_file_name in sa_jsons.keys():
            sa_jsons[sa_file_name].append(sa_obj)
        else:
            sa_jsons[sa_file_name] = [sa_obj]

    return sa_jsons, _create_classes(classes)



import json

from pathlib import Path
from PIL import Image
from tqdm import tqdm

from .coco_converter import CoCoConverter
from .coco_to_sa_pixel import (
    coco_instance_segmentation_to_sa_pixel,
    coco_panoptic_segmentation_to_sa_pixel
)
from .coco_to_sa_vector import (
    coco_instance_segmentation_to_sa_vector,
    coco_keypoint_detection_to_sa_vector, coco_object_detection_to_sa_vector
)
from .sa_pixel_to_coco import (
    sa_pixel_to_coco_instance_segmentation,
    sa_pixel_to_coco_panoptic_segmentation
)
from .sa_vector_to_coco import (
    sa_vector_to_coco_instance_segmentation,
    sa_vector_to_coco_keypoint_detection, sa_vector_to_coco_object_detection
)

from ....common import id2rgb, dump_output


class CocoPanopticConverterStrategy(CoCoConverter):
    name = "Panoptic converter"

    def __init__(self, args):
        super().__init__(args)
        self.__setup_conversion_algorithm()

    def __setup_conversion_algorithm(self):
        if self.direction == "to":
            if self.project_type == 'Pixel':
                self.conversion_algorithm = sa_pixel_to_coco_panoptic_segmentation
        else:
            self.conversion_algorithm = coco_panoptic_segmentation_to_sa_pixel

    def __str__(self, ):
        return '{} object'.format(self.name)

    def _sa_to_coco_single(self, id_, json_path, id_generator, cat_id_map):

        image_commons = self._prepare_single_image_commons(id_, json_path)
        res = self.conversion_algorithm(image_commons, id_generator, cat_id_map)

        return res

    def sa_to_output_format(self):
        out_json = self._create_skeleton()
        out_json['categories'] = self._create_categories(
            self.export_root / 'classes_mapper.json'
        )

        cat_id_map = json.load(open(self.export_root / 'classes_mapper.json'))

        images = []
        annotations = []
        id_generator = self._make_id_generator()

        jsons_gen = self.export_root.glob('*pixel.json')
        jsons = [path for path in jsons_gen]

        for id_, json_ in tqdm(enumerate(jsons, 1)):
            res = self._sa_to_coco_single(id_, json_, id_generator, cat_id_map)

            panoptic_mask = str(json_)[:-len('___pixel.json')] + '.png'

            Image.fromarray(id2rgb(res[2])).save(panoptic_mask)

            annotation = {
                'image_id': res[0]['id'],
                'file_name': panoptic_mask,
                'segments_info': res[1]
            }
            annotations.append(annotation)

            images.append(res[0])

        out_json['annotations'] = annotations
        out_json['images'] = images
        json_data = json.dumps(out_json, indent=4)
        with open(
            self.output_dir / '{}.json'.format(self.dataset_name), 'w+'
        ) as coco_json:

            coco_json.write(json_data)

        self.set_num_converted(len(jsons))

    def to_sa_format(self):
        json_data = self.export_root / (self.dataset_name + ".json")
        sa_classes = self._create_sa_classes(json_data)
        sa_jsons = self.conversion_algorithm(json_data, self.output_dir)
        dump_output(self.output_dir, self.platform, sa_classes, sa_jsons)


class CocoObjectDetectionStrategy(CoCoConverter):
    name = "ObjectDetection converter"

    def __init__(self, args):
        super().__init__(args)
        self.__setup_conversion_algorithm()

    def __setup_conversion_algorithm(self):

        if self.direction == "to":
            if self.project_type == 'Pixel':
                if self.task == 'instance_segmentation':
                    self.conversion_algorithm = sa_pixel_to_coco_instance_segmentation

            elif self.project_type == 'Vector':
                if self.task == 'instance_segmentation':
                    self.conversion_algorithm = sa_vector_to_coco_instance_segmentation
                elif self.task == 'object_detection':
                    self.conversion_algorithm = sa_vector_to_coco_object_detection
        else:
            if self.project_type == 'Pixel':
                if self.task == 'instance_segmentation':
                    self.conversion_algorithm = coco_instance_segmentation_to_sa_pixel
            elif self.project_type == 'Vector':
                if self.task == 'instance_segmentation':
                    self.conversion_algorithm = coco_instance_segmentation_to_sa_vector
                elif self.task == 'object_detection':
                    self.conversion_algorithm = coco_object_detection_to_sa_vector

    def __str__(self, ):
        return '{} object'.format(self.name)

    def _sa_to_coco_single(self, id_, json_path, id_generator, cat_id_map):

        image_commons = self._prepare_single_image_commons(id_, json_path)

        def make_annotation(
            category_id, image_id, bbox, segmentation, area, anno_id
        ):
            if self.task == 'object_detection':
                segmentation = [
                    [
                        bbox[0], bbox[1], bbox[0], bbox[1] + bbox[3],
                        bbox[0] + bbox[2], bbox[1] + bbox[3], bbox[0] + bbox[2],
                        bbox[1]
                    ]
                ]
            annotation = {
                'id': anno_id,  # making sure ids are unique
                'image_id': image_id,
                'segmentation': segmentation,
                'iscrowd': 0,
                'bbox': bbox,
                'area': area,
                'category_id': category_id
            }

            return annotation

        res = self.conversion_algorithm(
            make_annotation, image_commons, id_generator, cat_id_map
        )
        return res

    def sa_to_output_format(self):

        out_json = self._create_skeleton()
        out_json['categories'] = self._create_categories(
            self.export_root / 'classes_mapper.json'
        )

        cat_id_map = json.load(open(self.export_root / 'classes_mapper.json'))

        jsons = self._load_sa_jsons()
        images = []
        annotations = []
        id_generator = self._make_id_generator()
        for id_, json_ in tqdm(enumerate(jsons)):
            try:
                res = self._sa_to_coco_single(
                    id_, json_, id_generator, cat_id_map
                )
            except Exception as e:
                raise
            images.append(res[0])
            if len(res[1]) < 1:
                self.increase_converted_count()
            for ann in res[1]:
                annotations.append(ann)
        out_json['annotations'] = annotations
        out_json['images'] = images

        json_data = json.dumps(out_json, indent=4)
        with open(
            self.output_dir / '{}.json'.format(self.dataset_name), 'w+'
        ) as coco_json:
            coco_json.write(json_data)
        print("NUMBER OF IMAGES FAILED TO CONVERT", self.failed_conversion_cnt)

        self.set_num_converted(len(jsons))

    def to_sa_format(self):
        json_data = self.export_root / (self.dataset_name + ".json")
        sa_classes = self._create_sa_classes(json_data)
        sa_jsons = self.conversion_algorithm(json_data, self.output_dir)
        dump_output(self.output_dir, self.platform, sa_classes, sa_jsons)


class CocoKeypointDetectionStrategy(CoCoConverter):
    name = 'Keypoint Detection Converter'

    def __init__(self, args):
        super().__init__(args)
        self.__setup_conversion_algorithm()

    def __str__(self):
        return '{} object'.format(self.name)

    def __setup_conversion_algorithm(self):
        if self.direction == "to":
            self.conversion_algorithm = sa_vector_to_coco_keypoint_detection
        else:
            self.conversion_algorithm = coco_keypoint_detection_to_sa_vector

    def __make_image_info(self, json_path, id_, source_type):
        if source_type == 'Pixel':
            rm_len = len('___pixel.json')
        elif source_type == 'Vector':
            rm_len = len('___objects.json')

        img_width, img_height = 0, 0
        json_data = json.load(open(json_path))
        for annot in json_data:
            if 'type' in annot and annot['type'] == 'meta':
                img_height = annot['height']
                img_width = annot['width']

        image_path = str(json_path)[:-rm_len]

        if img_width == 0 and img_height == 0:
            img_width, img_height = Image.open(image_path).size

        image_info = {
            'id': id_,
            'file_name': Path(image_path).name,
            'height': img_height,
            'width': img_width,
            'license': 1
        }

        return image_info

    def sa_to_output_format(self):
        out_json = self._create_skeleton()
        jsons = self._load_sa_jsons()

        id_generator = self._make_id_generator()
        id_generator_anno = self._make_id_generator()
        id_generator_img = self._make_id_generator()
        res = self.conversion_algorithm(
            jsons, id_generator, id_generator_anno, id_generator_img,
            self.__make_image_info
        )

        out_json['categories'] = res[0]
        out_json['annotations'] = res[1]
        out_json['images'] = res[2]
        json_data = json.dumps(out_json, indent=4)

        with open(
            self.output_dir / '{}.json'.format(self.dataset_name), 'w+'
        ) as coco_json:
            coco_json.write(json_data)

        self.set_num_converted(len(out_json['images']))

    def to_sa_format(self):
        json_data = self.export_root / (self.dataset_name + ".json")
        sa_classes = self._create_sa_classes(json_data)
        sa_jsons = self.conversion_algorithm(json_data, self.output_dir)
        dump_output(self.output_dir, self.platform, sa_classes, sa_jsons)



import json

import cv2
import numpy as np

from pathlib import Path
from collections import defaultdict
from tqdm import tqdm

from superannotate.pycocotools_sa.coco import COCO
from superannotate.pycocotools_sa import mask as maskUtils

from ....common import id2rgb


def _rle_to_polygon(coco_json, annotation):
    coco = COCO(coco_json)
    binary_mask = coco.annToMask(annotation)
    contours, _ = cv2.findContours(
        binary_mask.astype(np.uint8), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    segmentation = []

    for contour in contours:
        contour = contour.flatten().tolist()
        if len(contour) > 4:
            segmentation.append(contour)
        if len(segmentation) == 0:
            continue
    return segmentation


def coco_instance_segmentation_to_sa_vector(coco_path, images_path):
    coco_json = json.load(open(coco_path))
    image_id_to_annotations = {}
    cat_id_to_cat = {}
    for cat in coco_json['categories']:
        cat_id_to_cat[cat['id']] = cat

    instance_groups = {}
    for annot in coco_json['annotations']:
        if 'id' in annot:
            if annot['id'] in instance_groups:
                instance_groups[annot['id']] += 1
            else:
                instance_groups[annot['id']] = 1

    for annot in tqdm(coco_json['annotations'], "Converting annotations"):
        if isinstance(annot['segmentation'], dict):
            if isinstance(annot['segmentation']['counts'], list):
                annot['segmentation'] = _rle_to_polygon(coco_path, annot)
            else:
                binary_mask = maskUtils.decode(annot['segmentation'])
                contours, _ = cv2.findContours(
                    binary_mask.astype(np.uint8), cv2.RETR_EXTERNAL,
                    cv2.CHAIN_APPROX_SIMPLE
                )
                segments = []
                for contour in contours:
                    contour = contour.flatten().tolist()
                    if len(contour) > 4:
                        segments.append(contour)
                annot['segmentation'] = segments

        cat = cat_id_to_cat[annot['category_id']]
        sa_polygon_loader = []
        groupid = 0
        for polygon in annot['segmentation']:
            if 'id' in annot:
                if annot['id'] in instance_groups and instance_groups[
                    annot['id']] > 1:
                    groupid = annot['id']
            sa_obj = {
                'type': 'polygon',
                'points': polygon,
                'className': cat['name'],
                'classId': cat['id'],
                'attributes': [],
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': groupid,
                'imageId': annot['image_id']
            }
            sa_polygon_loader.append(sa_obj)

        for polygon in sa_polygon_loader:
            if polygon['imageId'] not in image_id_to_annotations:
                image_id_to_annotations[polygon['imageId']] = [polygon]
            else:
                image_id_to_annotations[polygon['imageId']].append(polygon)

    sa_jsons = {}
    for img in tqdm(coco_json['images'], "Writing annotations to disk"):
        if 'file_name' in img:
            image_path = Path(img['file_name']).name
        else:
            image_path = img['coco_url'].split('/')[-1]

        if img['id'] not in image_id_to_annotations:
            f_loader = []
            # continue
        else:
            f_loader = image_id_to_annotations[img['id']]
        file_name = image_path + "___objects.json"
        sa_jsons[file_name] = f_loader
    return sa_jsons


def coco_object_detection_to_sa_vector(coco_path, images_path):
    coco_json = json.load(open(coco_path))
    image_id_to_annotations = {}
    cat_id_to_cat = {}
    for cat in coco_json['categories']:
        cat_id_to_cat[cat['id']] = cat

    instance_groups = {}
    for annot in coco_json['annotations']:
        if 'id' in annot:
            if annot['id'] in instance_groups:
                instance_groups[annot['id']] += 1
            else:
                instance_groups[annot['id']] = 1

    for annot in tqdm(coco_json['annotations'], "Converting annotations"):
        if isinstance(annot['segmentation'], dict):
            if isinstance(annot['segmentation']['counts'], list):
                annot['segmentation'] = _rle_to_polygon(coco_path, annot)
            else:
                binary_mask = maskUtils.decode(annot['segmentation'])
                contours, _ = cv2.findContours(
                    binary_mask.astype(np.uint8), cv2.RETR_EXTERNAL,
                    cv2.CHAIN_APPROX_SIMPLE
                )
                segments = []
                for contour in contours:
                    contour = contour.flatten().tolist()
                    if len(contour) > 4:
                        segments.append(contour)
                annot['segmentation'] = segments

        cat = cat_id_to_cat[annot['category_id']]
        groupid = 0
        for polygon in annot['segmentation']:
            if 'id' in annot:
                if annot['id'] in instance_groups and instance_groups[
                    annot['id']] > 1:
                    groupid = annot['id']

        sa_dict_bbox = {
            'type': 'bbox',
            'points':
                {
                    'x1': annot['bbox'][0],
                    'y1': annot['bbox'][1],
                    'x2': annot['bbox'][0] + annot['bbox'][2],
                    'y2': annot['bbox'][1] + annot['bbox'][3]
                },
            'className': cat['name'],
            'classId': cat['id'],
            'attributes': [],
            'probability': 100,
            'locked': False,
            'visible': True,
            'groupId': groupid,
            'imageId': annot['image_id']
        }

        if sa_dict_bbox['imageId'] not in image_id_to_annotations:
            image_id_to_annotations[sa_dict_bbox['imageId']] = [sa_dict_bbox]
        else:
            image_id_to_annotations[sa_dict_bbox['imageId']
                                   ].append(sa_dict_bbox)

    sa_jsons = {}
    for img in tqdm(coco_json['images'], "Writing annotations to disk"):
        if 'file_name' in img:
            image_path = Path(img['file_name']).name
        else:
            image_path = img['coco_url'].split('/')[-1]

        if img['id'] not in image_id_to_annotations:
            f_loader = []
            # continue
        else:
            f_loader = image_id_to_annotations[img['id']]
        file_name = image_path + "___objects.json"
        sa_jsons[file_name] = f_loader
    return sa_jsons


def coco_keypoint_detection_to_sa_vector(coco_path, images_path):
    coco_json = json.load(open(coco_path))
    loader = []

    cat_id_to_cat = {}
    for cat in coco_json["categories"]:
        cat_id_to_cat[cat["id"]] = {
            "name": cat["name"],
            "keypoints": cat["keypoints"],
            "skeleton": cat["skeleton"]
        }

    for annot in tqdm(coco_json['annotations'], 'Converting annotations'):
        if annot['num_keypoints'] > 0:
            sa_points = [
                item for index, item in enumerate(annot['keypoints'])
                if (index + 1) % 3 != 0
            ]

            sa_points = [
                (sa_points[i], sa_points[i + 1])
                for i in range(0, len(sa_points), 2)
            ]
            if annot["category_id"] in cat_id_to_cat.keys():
                keypoint_names = cat_id_to_cat[annot["category_id"]
                                              ]['keypoints']

                sa_template = {
                    'type': 'template',
                    'classId': annot['category_id'],
                    'probability': 100,
                    'points': [],
                    'connections': [],
                    'attributes': [],
                    'attributeNames': [],
                    'groupId': annot['id'],
                    'pointLabels': {},
                    'locked': False,
                    'visible': True,
                    'templateId': -1,
                    'className': cat_id_to_cat[annot["category_id"]]["name"],
                    'templateName': 'skeleton',
                    'imageId': annot['image_id']
                }

                bad_points = []
                id_mapping = {}
                index = 1
                for point_index, point in enumerate(sa_points):
                    if sa_points[point_index] == (0, 0):
                        bad_points.append(point_index + 1)
                        continue
                    id_mapping[point_index + 1] = index
                    sa_template['points'].append(
                        {
                            'id': index,
                            'x': point[0],
                            'y': point[1]
                        }
                    )
                    index += 1

                for connection in cat_id_to_cat[annot["category_id"]
                                               ]['skeleton']:

                    index = cat['skeleton'].index(connection)
                    from_point = cat_id_to_cat[annot["category_id"]
                                              ]['skeleton'][index][0]
                    to_point = cat_id_to_cat[annot["category_id"]
                                            ]['skeleton'][index][1]

                    if from_point in bad_points or to_point in bad_points:
                        continue

                    sa_template['connections'].append(
                        {
                            'id': index + 1,
                            'from': id_mapping[from_point],
                            'to': id_mapping[to_point]
                        }
                    )

                for kp_index, kp_name in enumerate(keypoint_names):
                    if kp_index + 1 in bad_points:
                        continue
                    sa_template['pointLabels'][id_mapping[kp_index + 1] -
                                               1] = kp_name

                for img in coco_json['images']:
                    if sa_template['imageId'] == img['id']:
                        loader.append((img['id'], sa_template))

        sa_jsons = {}
        for img in coco_json['images']:
            f_loader = []
            for img_id, img_data in loader:
                if img['id'] == img_id:
                    f_loader.append(img_data)
            file_name = str(Path(img['file_name']).name) + "___objects.json"
            sa_jsons[file_name] = f_loader
    return sa_jsons



import json

from .vott_converter import VoTTConverter
from .vott_to_sa_vector import vott_object_detection_to_sa_vector, vott_instance_segmentation_to_sa_vector, vott_to_sa

from ....common import dump_output


class VoTTObjectDetectionStrategy(VoTTConverter):
    name = "ObjectDetection converter"

    def __init__(self, args):
        super().__init__(args)
        self.__setup_conversion_algorithm()

    def __setup_conversion_algorithm(self):
        if self.direction == "from":
            if self.project_type == "Vector":
                if self.task == "object_detection":
                    self.conversion_algorithm = vott_object_detection_to_sa_vector
                elif self.task == 'instance_segmentation':
                    self.conversion_algorithm = vott_instance_segmentation_to_sa_vector
                elif self.task == 'vector_annotation':
                    self.conversion_algorithm = vott_to_sa

    def __str__(self):
        return '{} object'.format(self.name)

    def to_sa_format(self):
        json_data = self.get_file_list()
        sa_jsons, sa_classes = self.conversion_algorithm(json_data)
        dump_output(self.output_dir, self.platform, sa_classes, sa_jsons)



import cv2 as cv
import numpy as np

from ....pycocotools_sa import mask as cocomask


def __instance_object_commons_per_instance(
    instance, id_generator, image_commons, cat_id_map
):
    if "parts" not in instance:
        return None
    anno_id = next(id_generator)
    parts = [int(part["color"][1:], 16) for part in instance["parts"]]

    if instance['className'] in cat_id_map:
        category_id = cat_id_map[instance['className']]
    else:
        category_id = instance['classId']
    instance_bitmask = np.isin(image_commons.flat_mask, parts)

    databytes = instance_bitmask * np.uint8(255)
    contours, _ = cv.findContours(
        databytes, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_NONE
    )
    coco_instance_mask = cocomask.encode(np.asfortranarray(instance_bitmask))

    bbox = cocomask.toBbox(coco_instance_mask).tolist()
    area = int(cocomask.area(coco_instance_mask))
    return (bbox, area, contours, category_id, anno_id)


def instance_object_commons(image_commons, id_generator, cat_id_map):
    sa_ann_json = image_commons.sa_ann_json
    commons_lst = [
        __instance_object_commons_per_instance(
            x, id_generator, image_commons, cat_id_map
        ) for x in sa_ann_json
    ]
    return commons_lst


def sa_pixel_to_coco_instance_segmentation(
    make_annotation, image_commons, id_generator, cat_id_map
):
    commons_lst = instance_object_commons(
        image_commons, id_generator, cat_id_map
    )
    image_info = image_commons.image_info
    annotations_per_image = []
    for common in commons_lst:
        if common is None:
            continue
        bbox, area, contours, category_id, anno_id = common
        if category_id < 0:
            continue
        segmentation = [
            contour.flatten().tolist()
            for contour in contours if len(contour.flatten().tolist()) >= 5
        ]

        annotations_per_image.append(
            make_annotation(
                category_id, image_info['id'], bbox, segmentation, area, anno_id
            )
        )

    return (image_info, annotations_per_image)


def sa_pixel_to_coco_panoptic_segmentation(
    image_commons, id_generator, cat_id_map
):

    sa_ann_json = image_commons.sa_ann_json
    flat_mask = image_commons.flat_mask
    ann_mask = image_commons.ann_mask

    segments_info = []

    for instance in sa_ann_json:

        if 'parts' not in instance:
            continue

        parts = [int(part['color'][1:], 16) for part in instance['parts']]

        if instance['className'] in cat_id_map:
            category_id = cat_id_map[instance['className']]
        else:
            category_id = instance['classId']

        instance_bitmask = np.isin(flat_mask, parts)
        segment_id = next(id_generator)
        ann_mask[instance_bitmask] = segment_id
        coco_instance_mask = cocomask.encode(
            np.asfortranarray(instance_bitmask)
        )
        bbox = cocomask.toBbox(coco_instance_mask).tolist()
        area = int(cocomask.area(coco_instance_mask))

        segment_info = {
            'id': segment_id,
            'category_id': category_id,
            'area': area,
            'bbox': bbox,
            'iscrowd': 0
        }

        segments_info.append(segment_info)

    return (image_commons.image_info, segments_info, image_commons.ann_mask)



import os
import cv2
import json
import numpy as np
import xml.etree.ElementTree as ET
from tqdm import tqdm


def _generate_polygons(object_mask_path, class_mask_path):
    segmentation = []

    object_mask = cv2.imread(str(object_mask_path), cv2.IMREAD_GRAYSCALE)
    class_mask = cv2.imread(str(class_mask_path), cv2.IMREAD_GRAYSCALE)

    object_unique_colors = np.unique(object_mask)

    index = 1
    groupId = 0
    for unique_color in object_unique_colors:
        if unique_color == 0 or unique_color == 220:
            continue
        else:
            mask = np.zeros_like(object_mask)
            mask[object_mask == unique_color] = 255
            contours, _ = cv2.findContours(
                mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
            )

            segment = []
            if len(contours) > 1:
                for contour in contours:
                    segment.append(contour.flatten().tolist())
                groupId = index
                index += 1
            else:
                segment.append(contours[0].flatten().tolist())
                groupId = 0

            segmentation.append((segment, groupId))

    return segmentation


def _iou(bbox1, bbox2):
    xmin1, ymin1, xmax1, ymax1 = bbox1
    xmin2, ymin2, xmax2, ymax2 = bbox2

    x = max(0, min(xmax1, xmax2) - max(xmin1, xmin2))
    y = max(0, min(ymax1, ymax2) - max(ymin1, ymin2))
    return x * y / float(
        (xmax1 - xmin1) * (ymax1 - ymin1) + (xmax2 - xmin2) *
        (ymax2 - ymin2) - x * y
    )


def _generate_instances(polygon_instances, voc_instances):
    instances = []
    for polygon, group_id in polygon_instances:
        ious = []
        if len(polygon) > 1:
            temp = []
            for poly in polygon:
                temp += poly
        else:
            temp = polygon[0]
        bbox_poly = [
            min(temp[::2]),
            min(temp[1::2]),
            max(temp[::2]),
            max(temp[1::2])
        ]
        for class_name, bbox in voc_instances:
            ious.append(_iou(bbox_poly, bbox))
        ind = np.argmax(ious)
        for poly in polygon:
            instances.append(
                {
                    'className': voc_instances[ind][0],
                    'polygon': poly,
                    'bbox': voc_instances[ind][1],
                    'groupId': group_id
                }
            )
    return instances


def _get_voc_instances_from_xml(file_path):
    with open(os.path.splitext(file_path)[0] + ".xml") as f:
        tree = ET.parse(f)
    instances = tree.findall('object')
    voc_instances = []
    for instance in instances:
        class_name = instance.find("name").text
        bbox = instance.find("bndbox")
        bbox = [
            float(bbox.find(x).text) for x in ["xmin", "ymin", "xmax", "ymax"]
        ]
        voc_instances.append((class_name, bbox))
    return voc_instances


def _create_classes(classes):
    sa_classes = []
    for class_ in set(classes):
        color = np.random.choice(range(256), size=3)
        hexcolor = "#%02x%02x%02x" % tuple(color)
        sa_class = {"name": class_, "color": hexcolor, "attribute_groups": []}
        sa_classes.append(sa_class)
    return sa_classes


def voc_instance_segmentation_to_sa_vector(voc_root):
    # classes = {}
    classes = []
    object_masks_dir = voc_root / 'SegmentationObject'
    class_masks_dir = voc_root / 'SegmentationClass'
    annotation_dir = voc_root / "Annotations"

    file_list = object_masks_dir.glob('*')
    sa_jsons = {}
    for filename in tqdm(file_list):
        polygon_instances = _generate_polygons(
            object_masks_dir / filename.name, class_masks_dir / filename.name
        )
        voc_instances = _get_voc_instances_from_xml(
            annotation_dir / filename.name
        )
        maped_instances = _generate_instances(polygon_instances, voc_instances)
        sa_loader = []
        for instance in maped_instances:
            sa_polygon = {
                'type': 'polygon',
                'points': instance["polygon"],
                'className': instance["className"],
                'attributes': [],
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': instance['groupId']
            }
            sa_loader.append(sa_polygon)

            classes.append(instance["className"])

        sa_file_name = os.path.splitext(filename.name
                                       )[0] + ".jpg___objects.json"
        sa_jsons[sa_file_name] = sa_loader

    classes = _create_classes(classes)
    return (classes, sa_jsons, None)


def voc_object_detection_to_sa_vector(voc_root):
    classes = []
    annotation_dir = voc_root / "Annotations"
    files_list = annotation_dir.glob('*')
    sa_jsons = {}
    for filename in tqdm(files_list):
        voc_instances = _get_voc_instances_from_xml(
            annotation_dir / filename.name
        )
        sa_loader = []
        for class_name, bbox in voc_instances:
            classes.append(class_name)

            sa_bbox = {
                'type': "bbox",
                'className': class_name,
                'probability': 100,
                'points': [],
                'attributes': [],
                'points':
                    {
                        "x1": bbox[0],
                        "y1": bbox[1],
                        "x2": bbox[2],
                        "y2": bbox[3]
                    }
            }
            sa_loader.append(sa_bbox)

        sa_file_name = os.path.splitext(filename.name
                                       )[0] + ".jpg___objects.json"
        sa_jsons[sa_file_name] = sa_loader

    classes = _create_classes(classes)
    return (classes, sa_jsons, None)



import argparse
import json
import logging
import sys
from pathlib import Path
import tempfile

import superannotate as sa

from .exceptions import SABaseException

logger = logging.getLogger("superannotate-python-sdk")

_CLI_COMMAND = Path(sys.argv[0]).name


def ask_token():
    config_dir = Path.home() / ".superannotate"
    config_filename = "config.json"
    config_file = config_dir / config_filename
    if config_file.is_file():
        yes_no = input(
            f"File {config_file} exists. Do you want to overwrite? [y/n] : "
        )
        if yes_no != "y":
            return
    token = input(
        "Input the team SDK token from https://app.superannotate.com/team : "
    )
    config_dir.mkdir(exist_ok=True)
    if config_file.is_file():
        existing_config = json.load(open(config_file))
        existing_config["token"] = token
        json.dump(existing_config, open(config_file, "w"), indent=4)
        logger.info("Configuration file %s successfully updated.", config_file)
    else:
        json.dump(
            {
                "token": token,
                "main_endpoint": "https://api.annotate.online",
                "ssl_verify": True
            },
            open(config_file, "w"),
            indent=4
        )
        logger.info("Configuration file %s successfully created.", config_file)


def main():
    available_commands = "Available commands to superannotate CLI are: init version create-project upload-images upload-videos upload-preannotations upload-annotations export-project"
    if len(sys.argv) == 1:
        raise SABaseException(
            0, "No command given to superannotate CLI. " + available_commands
        )
    command = sys.argv[1]
    further_args = sys.argv[2:]

    if command == "create-project":
        create_project(command, further_args)
    elif command == "upload-images":
        image_upload(command, further_args)
    elif command == "upload-videos":
        video_upload(command, further_args)
    elif command in ["upload-preannotations", "upload-annotations"]:
        preannotations_upload(command, further_args)
    elif command == "init":
        ask_token()
    elif command == "export-project":
        export_project(command, further_args)
    elif command == "version":
        print(f"SuperAnnotate Python SDK version {sa.__version__}")
    else:
        raise SABaseException(
            0, "Wrong command " + command + " to superannotate CLI. " +
            available_commands
        )


def _list_str(values):
    return values.split(',')


def preannotations_upload(command_name, args):
    parser = argparse.ArgumentParser(prog=_CLI_COMMAND + " " + command_name)
    parser.add_argument(
        '--project', required=True, help='Project name to upload'
    )
    parser.add_argument(
        '--folder',
        required=True,
        help=
        'Folder (SuperAnnotate format) or JSON path (COCO format) from which to upload'
    )
    parser.add_argument(
        '--format',
        required=False,
        default="SuperAnnotate",
        help='Input preannotations format.'
    )
    parser.add_argument(
        '--dataset-name',
        required=False,
        help='Input annotations dataset name for COCO projects'
    )
    parser.add_argument(
        '--task',
        required=False,
        help=
        'Task type for COCO projects can be panoptic_segmentation (Pixel), instance_segmentation (Pixel), instance_segmentation (Vector), keypoint_detection (Vector)'
    )

    args = parser.parse_args(args)

    if args.format != "SuperAnnotate":
        if args.format != "COCO":
            raise sa.SABaseException(
                0, "Not supported annotations format " + args.format
            )
        if args.dataset_name is None:
            raise sa.SABaseException(
                0, "Dataset name should be present for COCO format upload."
            )
        if args.task is None:
            raise sa.SABaseException(
                0, "Task name should be present for COCO format upload."
            )

        logger.info("Annotations in format %s.", args.format)
        project_type = sa.get_project_metadata(args.project)["type"]

        tempdir = tempfile.TemporaryDirectory()
        tempdir_path = Path(tempdir.name)
        sa.import_annotation(
            args.folder, tempdir_path, "COCO", args.dataset_name, project_type,
            args.task
        )
        args.folder = tempdir_path

    sa.create_annotation_classes_from_classes_json(
        args.project,
        Path(args.folder) / "classes" / "classes.json"
    )

    if "pre" not in command_name:
        sa.upload_annotations_from_folder_to_project(
            project=args.project, folder_path=args.folder
        )
    else:
        sa.upload_preannotations_from_folder_to_project(
            project=args.project, folder_path=args.folder
        )


def create_project(command_name, args):
    parser = argparse.ArgumentParser(prog=_CLI_COMMAND + " " + command_name)
    parser.add_argument('--name', required=True, help='Project name to create')
    parser.add_argument(
        '--description', required=True, help='Project description'
    )
    parser.add_argument(
        '--type', required=True, help='Project type Vector or Pixel'
    )
    args = parser.parse_args(args)

    sa.create_project(args.name, args.description, args.type)


def video_upload(command_name, args):
    parser = argparse.ArgumentParser(prog=_CLI_COMMAND + " " + command_name)
    parser.add_argument(
        '--project', required=True, help='Project name to upload'
    )
    parser.add_argument(
        '--folder', required=True, help='Folder from which to upload'
    )
    parser.add_argument(
        '--recursive',
        default=False,
        action='store_true',
        help='Enables recursive subfolder upload.'
    )
    parser.add_argument(
        '--extensions',
        required=False,
        default=None,
        type=_list_str,
        help=
        'List of video extensions to include. Default is mp4,avi,mov,webm,flv,mpg,ogg'
    )
    parser.add_argument(
        '--set-annotation-status',
        required=False,
        default="NotStarted",
        help=
        'Set images\' annotation statuses after upload. Default is NotStarted'
    )
    parser.add_argument(
        '--target-fps',
        required=False,
        default=None,
        type=float,
        help=
        'How many frames per second need to extract from the videos (approximate).  If not specified all frames will be uploaded'
    )
    parser.add_argument(
        '--start-time',
        required=False,
        default=0.0,
        type=float,
        help=
        'Time (in seconds) from which to start extracting frames. Default is 0.0'
    )
    parser.add_argument(
        '--end-time',
        required=False,
        default=None,
        type=float,
        help=
        'Time (in seconds) up to which to extract frames. If it is not specified, then up to end'
    )
    args = parser.parse_args(args)

    sa.upload_videos_from_folder_to_project(
        project=args.project,
        folder_path=args.folder,
        extensions=args.extensions,
        annotation_status=args.set_annotation_status,
        recursive_subfolders=args.recursive,
        target_fps=args.target_fps,
        start_time=args.start_time,
        end_time=args.end_time
    )


def image_upload(command_name, args):
    parser = argparse.ArgumentParser(prog=_CLI_COMMAND + " " + command_name)
    parser.add_argument(
        '--project', required=True, help='Project name to upload'
    )
    parser.add_argument(
        '--folder', required=True, help='Folder from which to upload'
    )
    parser.add_argument(
        '--recursive',
        default=False,
        action='store_true',
        help='Enables recursive subfolder upload.'
    )
    parser.add_argument(
        '--extensions',
        default=None,
        type=_list_str,
        help=
        'List of image extensions to include. Default is jpg,jpeg,png,tif,tiff,webp,bmp'
    )
    parser.add_argument(
        '--set-annotation-status',
        required=False,
        default="NotStarted",
        help=
        'Set images\' annotation statuses after upload. Default is NotStarted'
    )
    args = parser.parse_args(args)

    sa.upload_images_from_folder_to_project(
        project=args.project,
        folder_path=args.folder,
        extensions=args.extensions,
        annotation_status=args.set_annotation_status,
        recursive_subfolders=args.recursive
    )


def export_project(command_name, args):
    parser = argparse.ArgumentParser(prog=_CLI_COMMAND + " " + command_name)
    parser.add_argument(
        '--project', required=True, help='Project name to export'
    )
    parser.add_argument(
        '--folder', required=True, help='Folder to which export'
    )
    parser.add_argument(
        '--include-fuse',
        default=False,
        action='store_true',
        help='Enables fuse image export'
    )
    parser.add_argument(
        '--disable-extract-zip-contents',
        default=False,
        action='store_true',
        help='Disables export zip extraction'
    )
    parser.add_argument(
        '--annotation-statuses',
        default=None,
        type=_list_str,
        help=
        'List of image annotation statuses to include in export. Default is InProgress,QualityCheck,Returned,Completed'
    )
    args = parser.parse_args(args)

    export = sa.prepare_export(
        args.project, args.annotation_statuses, args.include_fuse
    )
    sa.download_export(
        args.project, export, args.folder, not args.disable_extract_zip_contents
    )


if __name__ == "__main__":
    main()



import requests
import logging
from pathlib import Path
import cv2
import numpy as np

from ....common import hex_to_rgb, blue_color_generator

logger = logging.getLogger("superannotate-python-sdk")


def image_downloader(url, file_name):
    logger.info("Downloading mask for {}".format(file_name))
    r = requests.get(url, stream=True)
    with open(file_name, 'wb') as f:
        f.write(r.content)


def _create_classes_json(classes):
    sa_classes_loader = []
    for key, value in classes.items():
        sa_classes = {
            'name': key,
            'color': value['color'],
            'attribute_groups': []
        }
        attribute_groups = []
        for attr_group_key, attr_group in value['attribute_groups'].items():
            attr_loader = {
                'name': attr_group_key,
                'is_multiselect': attr_group['is_multiselect'],
                'attributes': []
            }
            for attr in attr_group['attributes']:
                attr_loader['attributes'].append({'name': attr})
            if attr_loader:
                attribute_groups.append(attr_loader)
        sa_classes['attribute_groups'] = attribute_groups

        sa_classes_loader.append(sa_classes)

    return sa_classes_loader


def _create_classes_id_map(json_data):
    classes = {}
    for d in json_data:
        if 'objects' not in d['Label'].keys():
            continue

        instances = d["Label"]["objects"]
        for instance in instances:
            class_name = instance["value"]
            if class_name not in classes.keys():
                color = instance["color"]
                classes[class_name] = {"color": color, 'attribute_groups': {}}

            if 'classifications' in instance.keys():
                classifications = instance['classifications']
                for classification in classifications:
                    if classification['value'] not in classes[class_name][
                        'attribute_groups']:
                        if 'answer' in classification.keys():
                            if isinstance(classification['answer'], str):
                                continue

                            classes[class_name]['attribute_groups'][
                                classification['value']] = {
                                    'is_multiselect': 0,
                                    'attributes': []
                                }
                            classes[class_name]['attribute_groups'][
                                classification['value']]['attributes'].append(
                                    classification['answer']['value']
                                )

                        elif 'answers' in classification.keys():
                            classes[class_name]['attribute_groups'][
                                classification['value']] = {
                                    'is_multiselect': 1,
                                    'attributes': []
                                }
                            for attr in classification['answers']:
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(attr['value'])

                    else:
                        if 'answer' in classification.keys():
                            classes[class_name]['attribute_groups'][
                                classification['value']]['attributes'].append(
                                    classification['answer']['value']
                                )
                        elif 'answers' in classification.keys():
                            for attr in classification['answers']:
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(attr['value'])
    return classes


def labelbox_instance_segmentation_to_sa_pixel(json_data):
    classes = _create_classes_id_map(json_data)
    sa_jsons = {}
    sa_masks = {}
    for d in json_data:
        file_name = d['External ID'] + '___pixel.json'
        mask_name = d['External ID'] + '___save.png'
        if 'objects' not in d['Label'].keys():
            sa_jsons[file_name] = []
            continue

        instances = d["Label"]["objects"]
        sa_loader = []
        blue_colors = blue_color_generator(len(instances))

        for i, instance in enumerate(instances):
            class_name = instance["value"]
            attributes = []
            if 'classifications' in instance.keys():
                classifications = instance['classifications']
                for classification in classifications:
                    group_name = classification['value']
                    if 'answer' in classification.keys(
                    ) and isinstance(classification['answer'], dict):
                        attribute_name = classification['answer']['value']
                        attr_dict = {
                            'name': attribute_name,
                            'groupName': group_name
                        }
                        attributes.append(attr_dict)
                    elif 'answers' in classification.keys():
                        for attr in classification['answers']:
                            attribute_name = attr['value']
                            attr_dict = {
                                'name': attribute_name,
                                'groupName': group_name
                            }
                            attributes.append(attr_dict)

            sa_obj = {
                'className': class_name,
                'attributes': attributes,
                'probability': 100,
                'locked': False,
                'visible': True,
                'parts': []
            }

            if 'bbox' in instance.keys() or 'polygon' in instance.keys(
            ) or 'line' in instance.keys() or 'point' in instance.keys():
                continue

            image_downloader(instance['instanceURI'], mask_name)
            mask = cv2.imread(mask_name)
            if i == 0:
                H, W, C = mask.shape
                sa_mask = np.zeros((H, W, C + 1))
            sa_mask[np.all(mask == [255, 255, 255], axis=2)
                   ] = list(hex_to_rgb(blue_colors[i]))[::-1] + [255]

            sa_obj['parts'].append({'color': blue_colors[i]})
            sa_loader.append(sa_obj.copy())
            Path(mask_name).unlink()

        sa_jsons[file_name] = sa_loader
        sa_masks[mask_name] = sa_mask

    sa_classes = _create_classes_json(classes)
    return sa_jsons, sa_classes, sa_masks



import logging

logger = logging.getLogger("superannotate-python-sdk")


class LabelBoxConverter():
    def __init__(self, args):
        logger.warning(
            "Beta feature. LabelBox to SuperAnnotate annotation format converter is in BETA state."
        )
        self.dataset_name = args.dataset_name
        self.project_type = args.project_type
        self.task = args.task
        self.output_dir = args.output_dir
        self.export_root = args.export_root
        self.direction = args.direction
        self.platform = args.platform



import logging
import os
from glob import glob

import cv2
import numpy as np

logger = logging.getLogger("superannotate-python-sdk")


def _create_classes(classes):
    classes_loader = []
    for id_, name in classes.items():
        color = np.random.choice(range(256), size=3)
        hexcolor = "#%02x%02x%02x" % tuple(color)
        sa_classes = {
            'id': id_ + 1,
            'name': name,
            'color': hexcolor,
            'attribute_groups': []
        }
        classes_loader.append(sa_classes)
    return classes_loader


def yolo_object_detection_to_sa_vector(data_path):
    classes = {}
    id_ = 0
    classes_file = open(data_path / 'classes.txt')
    for line in classes_file:
        key = line.rstrip()
        if key not in classes.keys():
            classes[id_] = key
            id_ += 1

    annotations = data_path.glob('*.txt')

    sa_jsons = {}
    for annotation in annotations:
        base_name = annotation.name
        if base_name == 'classes.txt':
            continue

        file = open(annotation)
        file_name = os.path.splitext(base_name)[0] + '.*'
        files_list = glob(os.path.join(data_path, file_name))
        if len(files_list) == 1:
            logger.warning(
                "'{}' image for annotation doesn't exist".format(annotation)
            )
            continue
        elif len(files_list) > 2:
            logger.warning(
                "'{}' multiple file for this annotation".format(annotation)
            )
            continue
        else:
            if os.path.splitext(files_list[0])[1] == '.txt':
                file_name = files_list[1]
            else:
                file_name = files_list[0]
        img = cv2.imread(file_name)
        H, W, _ = img.shape

        sa_loader = []
        for line in file:
            values = line.split()
            class_id = int(values[0])
            x_center = float(values[1]) * W
            y_center = float(values[2]) * H
            width = float(values[3]) * W
            height = float(values[4]) * H
            bbox = {
                'x1': x_center - width / 2,
                'y1': y_center - height / 2,
                'x2': x_center + width / 2,
                'y2': y_center + height / 2
            }
            sa_obj = {
                'type': 'bbox',
                'points': bbox,
                'className': classes[class_id],
                'attributes': [],
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': 0
            }
            sa_loader.append(sa_obj.copy())

        file_name = os.path.basename(file_name) + '___objects.json'
        sa_jsons[file_name] = sa_loader

    return sa_jsons, _create_classes(classes)



import json
import numpy as np
from ....common import rgb_to_hex


def _create_classes(class_id_map):
    sa_classes = []
    for key in class_id_map.keys():
        color = np.random.choice(range(256), size=3)
        hexcolor = "#%02x%02x%02x" % tuple(color)
        dd = {
            'id': class_id_map[key]['id'],
            'name': key,
            'color': hexcolor,
            'attribute_groups': []
        }
        for attributes, value in class_id_map[key]['attribute_groups'].items():
            attr_group = {
                'id': value['id'],
                'class_id': class_id_map[key]['id'],
                'name': attributes,
                'is_multiselect': value['is_multiselect'],
                'attributes': []
            }
            for attribute, attr_value in value['attributes'].items():
                attr = {
                    'id': attr_value,
                    'group_id': value['id'],
                    'name': attribute,
                    'groupName': attributes
                }
                attr_group['attributes'].append(attr.copy())
            dd['attribute_groups'].append(attr_group.copy())
        sa_classes.append(dd)
    return sa_classes


def vgg_object_detection_to_sa_vector(json_data, id_generator):
    sa_jsons = {}
    all_jsons = []
    for path in json_data:
        all_jsons.append(json.load(open(path)))

    class_id_map = {}
    for images in all_jsons:
        for key, img in images.items():
            file_name = img['filename'] + '___objects.json'
            sa_loader = []
            instances = img['regions']
            for instance in instances:
                if 'type' not in instance['region_attributes'].keys():
                    raise KeyError(
                        "'VGG' JSON should contain 'type' key which will be category name. Please correct JSON file."
                    )
                if not isinstance(instance['region_attributes']['type'], str):
                    raise ValueError(
                        "Wrong attribute was choosen for 'type' attribute."
                    )

                class_name = instance['region_attributes']['type']
                if class_name not in class_id_map.keys():
                    class_id_map[class_name] = {
                        'id': next(id_generator),
                        'attribute_groups': {}
                    }

                attr_keys = instance['region_attributes'].keys()
                attributes = []
                for attr in attr_keys:
                    if attr == 'type':
                        continue

                    if attr in class_id_map[class_name]['attribute_groups'
                                                       ].keys():
                        group_id = class_id_map[class_name]['attribute_groups'][
                            attr]['id']
                        group_name = attr
                        if isinstance(
                            instance['region_attributes'][attr], dict
                        ):
                            keys = instance['region_attributes'][attr].keys()
                        else:
                            keys = [instance['region_attributes'][attr]]

                        for key in keys:
                            if key not in class_id_map[class_name][
                                'attribute_groups'][attr]['attributes'].keys():
                                class_id_map[class_name]['attribute_groups'][
                                    attr]['attributes'][key] = next(
                                        id_generator
                                    )
                            dd = {
                                'id':
                                    class_id_map[class_name]['attribute_groups']
                                    [attr]['attributes'][key],
                                'name':
                                    key,
                                'groupId':
                                    group_id,
                                'groupName':
                                    group_name
                            }
                            attributes.append(dd.copy())

                    else:
                        multi = 0
                        if isinstance(
                            instance['region_attributes'][attr], dict
                        ):
                            multi = 1

                        class_id_map[class_name]['attribute_groups'][attr] = {
                            'id': next(id_generator),
                            'is_multiselect': multi,
                            'attributes': {}
                        }

                sa_obj = {
                    'className': class_name,
                    'classId': class_id_map[class_name]['id'],
                    'attributes': attributes,
                    'pointLabels': {},
                    'probability': 100,
                    'locked': False,
                    'visible': True,
                    'groupId': 0
                }
                if instance['shape_attributes']['name'] == 'polygon':
                    points = {
                        'x1': min(instance['shape_attributes']['all_points_x']),
                        'y1': min(instance['shape_attributes']['all_points_y']),
                        'x2': max(instance['shape_attributes']['all_points_x']),
                        'y2': max(instance['shape_attributes']['all_points_y'])
                    }
                    sa_obj['points'] = points
                    sa_obj['type'] = 'bbox'
                elif instance['shape_attributes']['name'] == 'rect':
                    points = {
                        'x1':
                            instance['shape_attributes']['x'],
                        'y1':
                            instance['shape_attributes']['y'],
                        'x2':
                            instance['shape_attributes']['x'] +
                            instance['shape_attributes']['width'],
                        'y2':
                            instance['shape_attributes']['y'] +
                            instance['shape_attributes']['height']
                    }
                    sa_obj['points'] = points
                    sa_obj['type'] = 'bbox'
                elif instance['shape_attributes']['name'] == 'ellipse':
                    if instance['shape_attributes']['rx'] > instance[
                        'shape_attributes']['ry']:
                        rx = instance['shape_attributes']['rx'] * np.cos(
                            instance['shape_attributes']['theta']
                        )
                        ry = max(
                            instance['shape_attributes']['rx'] *
                            np.sin(instance['shape_attributes']['theta']),
                            instance['shape_attributes']['ry']
                        )
                    else:
                        ry = instance['shape_attributes']['ry'] * np.sin(
                            instance['shape_attributes']['theta']
                        )
                        ry = max(
                            instance['shape_attributes']['ry'] *
                            np.cos(instance['shape_attributes']['theta']),
                            instance['shape_attributes']['rx']
                        )
                    points = {
                        'x1': instance['shape_attributes']['cx'] - rx,
                        'y1': instance['shape_attributes']['cy'] - ry,
                        'x2': instance['shape_attributes']['cx'] + rx,
                        'y2': instance['shape_attributes']['cy'] + ry
                    }
                    sa_obj['type'] = 'bbox'
                elif instance['shape_attributes']['name'] == 'circle':
                    points = {
                        'x1':
                            instance['shape_attributes']['cx'] -
                            instance['shape_attributes']['r'],
                        'y1':
                            instance['shape_attributes']['cy'] -
                            instance['shape_attributes']['r'],
                        'x2':
                            instance['shape_attributes']['cx'] +
                            instance['shape_attributes']['r'],
                        'y2':
                            instance['shape_attributes']['cy'] +
                            instance['shape_attributes']['r']
                    }
                    sa_obj['type'] = 'bbox'
                sa_obj['points'] = points

                sa_loader.append(sa_obj)
            sa_jsons[file_name] = sa_loader

        sa_classes = _create_classes(class_id_map)
    return sa_jsons, sa_classes


def vgg_instance_segmentation_to_sa_vector(json_data, id_generator):
    sa_jsons = {}
    all_jsons = []
    for path in json_data:
        all_jsons.append(json.load(open(path)))

    class_id_map = {}
    for images in all_jsons:
        for key, img in images.items():
            file_name = img['filename'] + '___objects.json'
            sa_loader = []
            instances = img['regions']
            for instance in instances:
                if instance['shape_attributes']['name'] != 'polygon':
                    continue

                if 'type' not in instance['region_attributes'].keys():
                    raise KeyError(
                        "'VGG' JSON should contain 'type' key which will be category name. Please correct JSON file."
                    )
                if not isinstance(instance['region_attributes']['type'], str):
                    raise ValueError(
                        "Wrong attribute was choosen for 'type' attribute."
                    )

                class_name = instance['region_attributes']['type']
                if class_name not in class_id_map.keys():
                    class_id_map[class_name] = {
                        'id': next(id_generator),
                        'attribute_groups': {}
                    }

                attr_keys = instance['region_attributes'].keys()
                attributes = []
                for attr in attr_keys:
                    if attr == 'type':
                        continue

                    if attr in class_id_map[class_name]['attribute_groups'
                                                       ].keys():
                        group_id = class_id_map[class_name]['attribute_groups'][
                            attr]['id']
                        group_name = attr
                        if isinstance(
                            instance['region_attributes'][attr], dict
                        ):
                            keys = instance['region_attributes'][attr].keys()
                        else:
                            keys = [instance['region_attributes'][attr]]

                        for key in keys:
                            if key not in class_id_map[class_name][
                                'attribute_groups'][attr]['attributes'].keys():
                                class_id_map[class_name]['attribute_groups'][
                                    attr]['attributes'][key] = next(
                                        id_generator
                                    )
                            dd = {
                                'id':
                                    class_id_map[class_name]['attribute_groups']
                                    [attr]['attributes'][key],
                                'name':
                                    key,
                                'groupId':
                                    group_id,
                                'groupName':
                                    group_name
                            }
                            attributes.append(dd.copy())

                    else:
                        multi = 0
                        if isinstance(
                            instance['region_attributes'][attr], dict
                        ):
                            multi = 1

                        class_id_map[class_name]['attribute_groups'][attr] = {
                            'id': next(id_generator),
                            'is_multiselect': multi,
                            'attributes': {}
                        }

                sa_obj = {
                    'className': class_name,
                    'classId': class_id_map[class_name]['id'],
                    'attributes': attributes,
                    'pointLabels': {},
                    'probability': 100,
                    'locked': False,
                    'visible': True,
                    'groupId': 0
                }
                points = []
                for x, y in zip(
                    instance['shape_attributes']['all_points_x'],
                    instance['shape_attributes']['all_points_y']
                ):
                    points.append(x)
                    points.append(y)
                sa_obj['points'] = points
                sa_obj['type'] = instance['shape_attributes']['name']

                sa_loader.append(sa_obj)
            sa_jsons[file_name] = sa_loader

        sa_classes = _create_classes(class_id_map)
    return sa_jsons, sa_classes


def vgg_to_sa(json_data, id_generator):
    sa_jsons = {}
    all_jsons = []
    for path in json_data:
        all_jsons.append(json.load(open(path)))

    class_id_map = {}
    for images in all_jsons:
        for key, img in images.items():
            file_name = img['filename'] + '___objects.json'
            sa_loader = []
            instances = img['regions']
            for instance in instances:
                if 'type' not in instance['region_attributes'].keys():
                    raise KeyError(
                        "'VGG' JSON should contain 'type' key which will be category name. Please correct JSON file."
                    )
                if not isinstance(instance['region_attributes']['type'], str):
                    raise ValueError(
                        "Wrong attribute was choosen for 'type' attribute."
                    )

                class_name = instance['region_attributes']['type']
                if class_name not in class_id_map.keys():
                    class_id_map[class_name] = {
                        'id': next(id_generator),
                        'attribute_groups': {}
                    }

                attr_keys = instance['region_attributes'].keys()
                attributes = []
                for attr in attr_keys:
                    if attr == 'type':
                        continue

                    if attr in class_id_map[class_name]['attribute_groups'
                                                       ].keys():
                        group_id = class_id_map[class_name]['attribute_groups'][
                            attr]['id']
                        group_name = attr
                        if isinstance(
                            instance['region_attributes'][attr], dict
                        ):
                            keys = instance['region_attributes'][attr].keys()
                        else:
                            keys = [instance['region_attributes'][attr]]

                        for key in keys:
                            if key not in class_id_map[class_name][
                                'attribute_groups'][attr]['attributes'].keys():
                                class_id_map[class_name]['attribute_groups'][
                                    attr]['attributes'][key] = next(
                                        id_generator
                                    )
                            dd = {
                                'id':
                                    class_id_map[class_name]['attribute_groups']
                                    [attr]['attributes'][key],
                                'name':
                                    key,
                                'groupId':
                                    group_id,
                                'groupName':
                                    group_name
                            }
                            attributes.append(dd.copy())

                    else:
                        multi = 0
                        if isinstance(
                            instance['region_attributes'][attr], dict
                        ):
                            multi = 1

                        class_id_map[class_name]['attribute_groups'][attr] = {
                            'id': next(id_generator),
                            'is_multiselect': multi,
                            'attributes': {}
                        }

                sa_obj = {
                    'className': class_name,
                    'classId': class_id_map[class_name]['id'],
                    'attributes': attributes,
                    'pointLabels': {},
                    'probability': 100,
                    'locked': False,
                    'visible': True,
                    'groupId': 0
                }
                if instance['shape_attributes']['name'] == 'polygon' or instance[
                    'shape_attributes']['name'] == 'polyline':
                    points = []
                    for x, y in zip(
                        instance['shape_attributes']['all_points_x'],
                        instance['shape_attributes']['all_points_y']
                    ):
                        points.append(x)
                        points.append(y)
                    sa_obj['points'] = points
                    sa_obj['type'] = instance['shape_attributes']['name']
                elif instance['shape_attributes']['name'] == 'rect':
                    points = {
                        'x1':
                            instance['shape_attributes']['x'],
                        'y1':
                            instance['shape_attributes']['y'],
                        'x2':
                            instance['shape_attributes']['x'] +
                            instance['shape_attributes']['width'],
                        'y2':
                            instance['shape_attributes']['y'] +
                            instance['shape_attributes']['height']
                    }
                    sa_obj['points'] = points
                    sa_obj['type'] = 'bbox'
                elif instance['shape_attributes']['name'] == 'ellipse':
                    sa_obj['type'] = 'ellipse'
                    sa_obj['cx'] = instance['shape_attributes']['cx']
                    sa_obj['cy'] = instance['shape_attributes']['cy']
                    sa_obj['rx'] = instance['shape_attributes']['rx']
                    sa_obj['ry'] = instance['shape_attributes']['ry']
                    sa_obj['angle'] = instance['shape_attributes']['theta']
                elif instance['shape_attributes']['name'] == 'circle':
                    sa_obj['type'] = 'ellipse'
                    sa_obj['cx'] = instance['shape_attributes']['cx']
                    sa_obj['cy'] = instance['shape_attributes']['cy']
                    sa_obj['rx'] = instance['shape_attributes']['r']
                    sa_obj['ry'] = instance['shape_attributes']['r']
                    sa_obj['angle'] = 0
                elif instance['shape_attributes']['name'] == 'point':
                    sa_obj['type'] = 'point'
                    sa_obj['x'] = instance['shape_attributes']['cx']
                    sa_obj['y'] = instance['shape_attributes']['cy']

                sa_loader.append(sa_obj)
            sa_jsons[file_name] = sa_loader

        sa_classes = _create_classes(class_id_map)
    return sa_jsons, sa_classes


from pathlib import Path
import logging

logger = logging.getLogger("superannotate-python-sdk")


class VGGConverter(object):
    def __init__(self, args):
        logger.warning(
            "Beta feature. VGG to SuperAnnotate annotation format converter is in BETA state."
        )
        self.dataset_name = args.dataset_name
        self.project_type = args.project_type
        self.task = args.task
        self.output_dir = args.output_dir
        self.export_root = args.export_root
        self.direction = args.direction
        self.platform = args.platform

    def get_file_list(self):
        json_file_list = []
        path = Path(self.export_root)
        if self.dataset_name != '':
            json_file_list.append(path.joinpath(self.dataset_name + '.json'))
        else:
            file_generator = path.glob('*.json')
            for gen in file_generator:
                json_file_list.append(gen)

        return json_file_list



"""
Main module for input converters
"""
from argparse import Namespace
from pathlib import Path

from ..exceptions import SABaseException
from .export_from_sa_conversions import export_from_sa
from .import_to_sa_conversions import import_to_sa
from .sa_conversion import (
    sa_convert_platform, sa_convert_project_type, split_coco
)

AVAILABLE_PLATFORMS = ["Desktop", "Web"]

ALLOWED_TASK_TYPES = [
    'panoptic_segmentation', 'instance_segmentation', 'keypoint_detection',
    'object_detection', 'vector_annotation'
]

ALLOWED_PROJECT_TYPES = ['Pixel', 'Vector']

ALLOWED_ANNOTATION_IMPORT_FORMATS = {
    'COCO':
        [
            ('Pixel', 'panoptic_segmentation'),
            ('Pixel', 'instance_segmentation'),
            ('Vector', 'keypoint_detection'),
            ('Vector', 'instance_segmentation'), ('Vector', 'object_detection')
        ],
    'VOC':
        [
            ('Vector', 'object_detection'), ('Vector', 'instance_segmentation'),
            ('Pixel', 'instance_segmentation')
        ],
    'LabelBox':
        [
            ('Vector', 'object_detection'), ('Vector', 'instance_segmentation'),
            ('Vector', 'vector_annotation'), ('Pixel', 'instance_segmentation')
        ],
    'DataLoop':
        [
            ('Vector', 'object_detection'), ('Vector', 'instance_segmentation'),
            ('Vector', 'vector_annotation')
        ],
    'Supervisely':
        [
            ('Vector', 'vector_annotation'), ('Vector', 'object_detection'),
            ('Vector', 'instance_segmentation'),
            ('Pixel', 'instance_segmentation'),
            ('Vector', 'keypoint_detection')
        ],
    'VoTT':
        [
            ('Vector', 'object_detection'), ('Vector', 'instance_segmentation'),
            ('Vector', 'vector_annotation')
        ],
    'SageMaker':
        [('Pixel', 'instance_segmentation'), ('Vector', 'object_detection')],
    'VGG':
        [
            ('Vector', 'object_detection'), ('Vector', 'instance_segmentation'),
            ('Vector', 'vector_annotation')
        ],
    'GoogleCloud': [('Vector', 'object_detection')],
    'YOLO': [('Vector', 'object_detection')]
}

ALLOWED_ANNOTATION_EXPORT_FORMATS = {
    'COCO':
        [
            ('Pixel', 'panoptic_segmentation'),
            ('Pixel', 'instance_segmentation'),
            ('Vector', 'instance_segmentation'),
            ('Vector', 'keypoint_detection'), ('Vector', 'object_detection')
        ]
}


def _passes_sanity_checks(args):
    if not isinstance(args.input_dir, (str, Path)):
        raise SABaseException(
            0, "'input_dir' should be 'str' or 'Path' type, not '%s'" %
            (type(args.input_dir))
        )

    if not isinstance(args.output_dir, (str, Path)):
        raise SABaseException(
            0, "'output_dir' should be 'str' or 'Path' type, not {}".format(
                type(args.output_dir)
            )
        )

    if args.dataset_format not in ALLOWED_ANNOTATION_IMPORT_FORMATS.keys():
        raise SABaseException(
            0, "'%s' converter doesn't exist. Possible candidates are '%s'" %
            (args.dataset_format, ALLOWED_ANNOTATION_IMPORT_FORMATS.keys())
        )

    if not isinstance(args.dataset_name, str):
        raise SABaseException(
            0, "'dataset_name' should be 'str' type, not {}".format(
                type(args.dataset_name)
            )
        )

    if args.project_type not in ALLOWED_PROJECT_TYPES:
        raise SABaseException(
            0, "Please enter valid project type: 'Pixel' or 'Vector'"
        )

    if args.task not in ALLOWED_TASK_TYPES:
        raise SABaseException(
            0, "Please enter valid task '%s'" % (ALLOWED_TASK_TYPES)
        )

    if 'platform' in args:
        if args.platform not in AVAILABLE_PLATFORMS:
            raise SABaseException(
                0, "Please enter valid platform: 'Desktop' or 'Web'"
            )

    if args.project_type == "Pixel" and args.platform == "Desktop":
        raise SABaseException(
            0,
            "Sorry, but Desktop Application doesn't support 'Pixel' projects."
        )


def _passes_converter_sanity(args, direction):
    converter_values = (args.project_type, args.task)
    test_passed = False
    if direction == 'import':
        if converter_values in ALLOWED_ANNOTATION_IMPORT_FORMATS[
            args.dataset_format]:
            test_passed = True
    else:
        if converter_values in ALLOWED_ANNOTATION_EXPORT_FORMATS[
            args.dataset_format]:
            test_passed = True

    if not test_passed:
        raise SABaseException(
            0,
            "Please enter valid converter values. You can check available candidates in the documentation (https://superannotate.readthedocs.io/en/stable/index.html)."
        )


def export_annotation(
    input_dir,
    output_dir,
    dataset_format,
    dataset_name,
    project_type="Vector",
    task="object_detection",
    platform="Web",
):
    """Converts SuperAnnotate annotation formate to the other annotation formats. Currently available (project_type, task) combinations for converter
    presented below:

    ==============  ======================
             From SA to COCO
    --------------------------------------
     project_type           task
    ==============  ======================
    Pixel           panoptic_segmentation
    Pixel           instance_segmentation
    Vector          instance_segmentation
    Vector          object_detection
    Vector          keypoint_detection
    ==============  ======================

    :param input_dir: Path to the dataset folder that you want to convert.
    :type input_dir: str
    :param output_dir: Path to the folder, where you want to have converted dataset.
    :type output_dir: str
    :param dataset_format: One of the formats that are possible to convert. Available candidates are: ["COCO"]
    :type dataset_format: str
    :param dataset_name: Will be used to create json file in the output_dir.
    :type dataset_name: str
    :param project_type: SuperAnnotate project type is either 'Vector' or 'Pixel' (Default: 'Vector')
                         'Vector' project creates <image_name>___objects.json for each image.
                         'Pixel' project creates <image_name>___pixel.jsons and <image_name>___save.png annotation mask for each image.
    :type project_type: str
    :param task: Task can be one of the following: ['panoptic_segmentation', 'instance_segmentation',
                 'keypoint_detection', 'object_detection']. (Default: "object_detection").
                 'keypoint_detection' can be used to converts keypoints from/to available annotation format.
                 'panoptic_segmentation' will use panoptic mask for each image to generate bluemask for SuperAnnotate annotation format and use bluemask to generate panoptic mask for invert conversion. Panoptic masks should be in the input folder.
                 'instance_segmentation' 'Pixel' project_type converts instance masks and 'Vector' project_type generates bounding boxes and polygons from instance masks. Masks should be in the input folder if it is 'Pixel' project_type.
                 'object_detection' converts objects from/to available annotation format
    :type task: str
    :param platform: SuperAnnotate has both 'Web' and 'Desktop' platforms. Choose from which one you are converting. (Default: "Web")
    :type platform: str

    """

    if isinstance(input_dir, str):
        input_dir = Path(input_dir)
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)

    args = Namespace(
        input_dir=input_dir,
        output_dir=output_dir,
        dataset_format=dataset_format,
        dataset_name=dataset_name,
        project_type=project_type,
        task=task,
        platform=platform,
    )

    _passes_sanity_checks(args)
    _passes_converter_sanity(args, 'export')

    export_from_sa(args)


def import_annotation(
    input_dir,
    output_dir,
    dataset_format,
    dataset_name='',
    project_type="Vector",
    task="object_detection",
    platform="Web",
    images_root='',
    images_extensions=['jpg']
):
    """Converts other annotation formats to SuperAnnotate annotation format. Currently available (project_type, task) combinations for converter
    presented below:

    ==============  ======================
             From COCO to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Pixel           panoptic_segmentation
    Pixel           instance_segmentation
    Vector          instance_segmentation
    Vector          object_detection
    Vector          keypoint_detection
    ==============  ======================

    ==============  ======================
             From VOC to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Pixel           instance_segmentation
    Vector          instance_segmentation
    Vector          object_detection
    ==============  ======================

    ==============  ======================
           From LabelBox to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    Vector          instance_segmentation
    Vector          vector_annotation
    ==============  ======================

    ==============  ======================
           From DataLoop to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    Vector          instance_segmentation
    Vector          vector_annotation
    ==============  ======================

    ==============  ======================
           From Supervisely to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    Vector          keypoint_detection
    Vector          vector_annotation
    Vector          instance_segmentation
    Pixel           instance_segmentation
    ==============  ======================

    ==============  ======================
           From VoTT to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          instance_segmentation
    Vector          object_detection
    Vector          vector_annotation
    ==============  ======================

    ==============  ======================
           From SageMaker to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Pixel           instance_segmentation
    Vector          objcet_detection
    ==============  ======================

    ==============  ======================
           From VGG to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          instance_segmentation
    Vector          object_detection
    Vector          vector_annotation
    ==============  ======================

    ==============  ======================
           From GoogleCloud to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    ==============  ======================

    ==============  ======================
           From YOLO to SA
    --------------------------------------
     project_type           task
    ==============  ======================
    Vector          object_detection
    ==============  ======================

    :param input_dir: Path to the dataset folder that you want to convert.
    :type input_dir: str
    :param output_dir: Path to the folder, where you want to have converted dataset.
    :type output_dir: str
    :param dataset_format: Annotation format to convert SuperAnnotate annotation format. Available candidates are: ["COCO", "VOC", "LabelBox", "DataLoop",
                        "Supervisely", 'VGG', 'YOLO', 'SageMake', 'VoTT', 'GoogleCloud']
    :type dataset_format: str
    :param dataset_name: Name of the json file in the input_dir, which should be converted.
    :type dataset_name: str
    :param project_type: SuperAnnotate project type is either 'Vector' or 'Pixel' (Default: 'Vector')
                         'Vector' project creates <image_name>___objects.json for each image.
                         'Pixel' project creates <image_name>___pixel.jsons and <image_name>___save.png annotation mask for each image.
    :type project_type: str
    :param task: Task can be one of the following: ['panoptic_segmentation', 'instance_segmentation',
                 'keypoint_detection', 'object_detection', 'vector_annotation']. (Default: "object_detection").
                 'keypoint_detection' can be used to converts keypoints from/to available annotation format.
                 'panoptic_segmentation' will use panoptic mask for each image to generate bluemask for SuperAnnotate annotation format and use bluemask to generate panoptic mask for invert conversion. Panoptic masks should be in the input folder.
                 'instance_segmentation' 'Pixel' project_type converts instance masks and 'Vector' project_type generates bounding boxes and polygons from instance masks. Masks should be in the input folder if it is 'Pixel' project_type.
                 'object_detection' converts objects from/to available annotation format
                 'vector_annotation' can be used to convert all annotations (point, ellipse, circule, cuboid and etc) to SuperAnnotate vector project.
    :param platform: SuperAnnotate has both 'Web' and 'Desktop' platforms. Choose to which platform you want convert. (Default: "Web")
    :type platform: str
    :param images_root: Additonal path to images directory in input_dir
    :type platform: str
    :param images_root: Additonal path to images directory in input_dir
    :type platform: str
    :param image_extensions: List of image files xtensions in the images_root folder
    :type platform: list

    """

    if isinstance(input_dir, str):
        input_dir = Path(input_dir)
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)

    args = Namespace(
        input_dir=input_dir,
        output_dir=output_dir,
        dataset_format=dataset_format,
        dataset_name=dataset_name,
        project_type=project_type,
        task=task,
        platform=platform,
        images_root=images_root,
        images_extensions=images_extensions
    )

    _passes_sanity_checks(args)
    _passes_converter_sanity(args, 'import')

    import_to_sa(args)


def convert_platform(input_dir, output_dir, input_platform):
    """ Converts SuperAnnotate input file structure from one platform too another.

    :param input_dir: Path to the dataset folder that you want to convert.
    :type input_dir: str or PathLike
    :param output_dir: Path to the folder where you want to have converted files.
    :type output_dir: str or PathLike
    :param input_platform: Original platform format type
    :type input_platform: str

    """
    param_info = [
        (input_dir, 'input_dir', (str, Path)),
        (output_dir, 'output_dir', (str, Path)),
        (input_platform, 'input_platform', str),
    ]
    for param in param_info:
        _type_sanity(param[0], param[1], param[2])

    if input_platform not in AVAILABLE_PLATFORMS:
        raise SABaseException(
            0, "Please enter valid platform: 'Desktop' or 'Web'"
        )

    if isinstance(input_dir, str):
        input_dir = Path(input_dir)
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)

    sa_convert_platform(input_dir, output_dir, input_platform)


def convert_project_type(input_dir, output_dir):
    """ Converts SuperAnnotate 'Vector' project type to 'Pixel' or reverse.

    :param input_dir: Path to the dataset folder that you want to convert.
    :type input_dir: str or PathLike
    :param output_dir: Path to the folder where you want to have converted files.
    :type output_dir: str or PathLike

    """
    param_info = [
        (input_dir, 'input_dir', (str, Path)),
        (output_dir, 'output_dir', (str, Path)),
    ]
    for param in param_info:
        _type_sanity(param[0], param[1], param[2])

    if isinstance(input_dir, str):
        input_dir = Path(input_dir)
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)

    sa_convert_project_type(input_dir, output_dir)


def coco_split_dataset(
    coco_json_path, image_dir, output_dir, dataset_list_name, ratio_list
):
    """ Splits COCO dataset to few datsets.

    :param coco_json_path: Path to main COCO JSON dataset, which should be splitted.
    :type coco_json_path: str or PathLike
    :param image_dir: Path to all images in the original dataset.
    :type coco_json_path: str or PathLike
    :param coco_json_path: Path to the folder where you want to output splitted COCO JSON files.
    :type coco_json_path: str or PathLike
    :param dataset_list_name: List of dataset names.
    :type dataset_list_name: list
    :param ratio_list: List of ratios for each splitted dataset.
    :type ratio_list: list
    """
    param_info = [
        (coco_json_path, 'coco_json_path', (str, Path)),
        (image_dir, 'image_dir', (str, Path)),
        (output_dir, 'output_dir', (str, Path)),
        (dataset_list_name, 'dataset_list_name', list),
        (ratio_list, 'ratio_list', list)
    ]
    for param in param_info:
        _type_sanity(param[0], param[1], param[2])

    for dataset_name in dataset_list_name:
        if not isinstance(dataset_name, (str, Path)):
            raise SABaseException(
                0,
                "'dataset_list_name' member should be 'str' or 'Path' type, not '%s'"
                % (type(dataset_name))
            )

    for ratio in ratio_list:
        if not isinstance(ratio, (int, float)):
            raise SABaseException(
                0,
                "'ratio_list' member should be 'int' or 'float' type, not '%s'"
                % (type(ratio))
            )

    if sum(ratio_list) != 100:
        raise SABaseException(0, "Sum of 'ratio_list' members must be '100'")

    if len(dataset_list_name) != len(ratio_list):
        raise SABaseException(
            0, "'dataset_list_name' and 'ratio_list' should have same lenght"
        )

    if isinstance(image_dir, str):
        image_dir = Path(image_dir)
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)

    split_coco(
        coco_json_path, image_dir, output_dir, dataset_list_name, ratio_list
    )


def _type_sanity(var, var_name, var_type):
    if not isinstance(var, var_type):
        raise SABaseException(
            0, "'{}' should be '{}' type, not '{}'".format(
                var_name, var_type, type(var)
            )
        )



def _create_classes_json(classes):
    sa_classes_loader = []
    for key, value in classes.items():
        sa_classes = {
            'name': key,
            'color': value['color'],
            'attribute_groups': []
        }
        attribute_groups = []
        for attr_group_key, attr_group in value['attribute_groups'].items():
            attr_loader = {
                'name': attr_group_key,
                'is_multiselect': attr_group['is_multiselect'],
                'attributes': []
            }
            for attr in attr_group['attributes']:
                attr_loader['attributes'].append({'name': attr})
            if attr_loader:
                attribute_groups.append(attr_loader)
        sa_classes['attribute_groups'] = attribute_groups

        sa_classes_loader.append(sa_classes)

    return sa_classes_loader


def _create_classes_id_map(json_data):
    classes = {}
    for d in json_data:
        if 'objects' not in d['Label'].keys():
            continue

        instances = d["Label"]["objects"]
        for instance in instances:
            class_name = instance["value"]
            if class_name not in classes.keys():
                color = instance["color"]
                classes[class_name] = {"color": color, 'attribute_groups': {}}

            if 'classifications' in instance.keys():
                classifications = instance['classifications']
                for classification in classifications:
                    if classification['value'] not in classes[class_name][
                        'attribute_groups']:
                        if 'answer' in classification.keys():
                            if isinstance(classification['answer'], str):
                                continue

                            classes[class_name]['attribute_groups'][
                                classification['value']] = {
                                    'is_multiselect': 0,
                                    'attributes': []
                                }
                            classes[class_name]['attribute_groups'][
                                classification['value']]['attributes'].append(
                                    classification['answer']['value']
                                )

                        elif 'answers' in classification.keys():
                            classes[class_name]['attribute_groups'][
                                classification['value']] = {
                                    'is_multiselect': 1,
                                    'attributes': []
                                }
                            for attr in classification['answers']:
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(attr['value'])

                    else:
                        if 'answer' in classification.keys():
                            classes[class_name]['attribute_groups'][
                                classification['value']]['attributes'].append(
                                    classification['answer']['value']
                                )
                        elif 'answers' in classification.keys():
                            for attr in classification['answers']:
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(attr['value'])
    return classes


def labelbox_object_detection_to_sa_vector(json_data):
    classes = _create_classes_id_map(json_data)
    sa_jsons = {}
    for d in json_data:
        if 'objects' not in d['Label'].keys():
            file_name = d['External ID'] + '___objects.json'
            sa_jsons[file_name] = []
            continue

        instances = d["Label"]["objects"]
        sa_loader = []
        for instance in instances:
            class_name = instance["value"]
            attributes = []
            if 'classifications' in instance.keys():
                classifications = instance['classifications']
                for classification in classifications:
                    group_name = classification['value']
                    if 'answer' in classification.keys(
                    ) and isinstance(classification['answer'], dict):
                        attribute_name = classification['answer']['value']
                        attr_dict = {
                            'name': attribute_name,
                            'groupName': group_name
                        }
                        attributes.append(attr_dict)
                    elif 'answers' in classification.keys():
                        for attr in classification['answers']:
                            attribute_name = attr['value']
                            attr_dict = {
                                'name': attribute_name,
                                'groupName': group_name
                            }
                            attributes.append(attr_dict)
            sa_obj = {
                'className': class_name,
                'attributes': attributes,
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': 0,
            }

            if 'bbox' in instance.keys():
                x1 = instance['bbox']['left']
                x2 = instance['bbox']['left'] + instance['bbox']['width']
                y1 = instance['bbox']['top']
                y2 = instance['bbox']['top'] + instance['bbox']['height']
                sa_obj['points'] = {'x1': x1, 'x2': x2, 'y1': y1, 'y2': y2}
                sa_obj['type'] = 'bbox'
                sa_loader.append(sa_obj)

        file_name = d['External ID'] + '___objects.json'
        sa_jsons[file_name] = sa_loader

    sa_classes = _create_classes_json(classes)
    return sa_jsons, sa_classes, None


def labelbox_instance_segmentation_to_sa_vector(json_data):
    classes = _create_classes_id_map(json_data)
    sa_jsons = {}
    for d in json_data:
        if 'objects' not in d['Label'].keys():
            file_name = d['External ID'] + '___objects.json'
            sa_jsons[file_name] = []
            continue

        instances = d["Label"]["objects"]
        sa_loader = []
        for instance in instances:
            class_name = instance["value"]
            attributes = []
            if 'classifications' in instance.keys():
                classifications = instance['classifications']
                for classification in classifications:
                    group_name = classification['value']
                    if 'answer' in classification.keys(
                    ) and isinstance(classification['answer'], dict):
                        attribute_name = classification['answer']['value']
                        attr_dict = {
                            'name': attribute_name,
                            'groupName': group_name
                        }
                        attributes.append(attr_dict)
                    elif 'answers' in classification.keys():
                        for attr in classification['answers']:
                            attribute_name = attr['value']
                            attr_dict = {
                                'name': attribute_name,
                                'groupName': group_name
                            }
                            attributes.append(attr_dict)

            sa_obj = {
                'className': class_name,
                'attributes': attributes,
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': 0,
            }

            if 'polygon' in instance.keys():
                sa_obj['type'] = 'polygon'
                points = []
                for point in instance['polygon']:
                    points.append(point['x'])
                    points.append(point['y'])
                sa_obj['points'] = points
                sa_loader.append(sa_obj)

        file_name = d['External ID'] + '___objects.json'
        sa_jsons[file_name] = sa_loader

    sa_classes = _create_classes_json(classes)
    return sa_jsons, sa_classes, None


def labelbox_to_sa(json_data):
    classes = _create_classes_id_map(json_data)
    sa_jsons = {}
    for d in json_data:
        if 'objects' not in d['Label'].keys():
            file_name = d['External ID'] + '___objects.json'
            sa_jsons[file_name] = []
            continue

        instances = d["Label"]["objects"]
        sa_loader = []
        for instance in instances:
            class_name = instance["value"]
            attributes = []
            if 'classifications' in instance.keys():
                classifications = instance['classifications']
                for classification in classifications:
                    group_name = classification['value']
                    if 'answer' in classification.keys(
                    ) and isinstance(classification['answer'], dict):
                        attribute_name = classification['answer']['value']
                        attr_dict = {
                            'name': attribute_name,
                            'groupName': group_name
                        }
                        attributes.append(attr_dict)
                    elif 'answers' in classification.keys():
                        for attr in classification['answers']:
                            attribute_name = attr['value']
                            attr_dict = {
                                'name': attribute_name,
                                'groupName': group_name
                            }
                            attributes.append(attr_dict)

            sa_obj = {
                'className': class_name,
                'attributes': attributes,
                'probability': 100,
                'locked': False,
                'visible': True,
                'groupId': 0,
            }
            if 'polygon' in instance.keys():
                sa_obj['type'] = 'polygon'
                points = []
                for point in instance['polygon']:
                    points.append(point['x'])
                    points.append(point['y'])
                sa_obj['points'] = points
                sa_loader.append(sa_obj)
            elif 'bbox' in instance.keys():
                x1 = instance['bbox']['left']
                x2 = instance['bbox']['left'] + instance['bbox']['width']
                y1 = instance['bbox']['top']
                y2 = instance['bbox']['top'] + instance['bbox']['height']
                sa_obj['points'] = {'x1': x1, 'x2': x2, 'y1': y1, 'y2': y2}
                sa_obj['type'] = 'bbox'
                sa_loader.append(sa_obj)
            elif 'line' in instance.keys():
                sa_obj['type'] = 'polyline'
                points = []
                for point in instance['line']:
                    points.append(point['x'])
                    points.append(point['y'])
                sa_obj['points'] = points
                sa_loader.append(sa_obj)
            elif 'point' in instance.keys():
                sa_obj['points'] = {
                    'x': instance['point']['x'],
                    'y': instance['point']['y']
                }
                sa_obj['type'] = 'point'
                sa_loader.append(sa_obj)

        file_name = d['External ID'] + '___objects.json'
        sa_jsons[file_name] = sa_loader

    sa_classes = _create_classes_json(classes)
    return sa_jsons, sa_classes, None